# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2012 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
gettext for openstack-common modules.

Usual usage in an openstack.common module:

    from savanna.openstack.common.gettextutils import _
"""

import gettext
import os

_localedir = os.environ.get('savanna'.upper() + '_LOCALEDIR')
_t = gettext.translation('savanna', localedir=_localedir, fallback=True)


def _(msg):
    return _t.ugettext(msg)


def install(domain):
    """Install a _() function using the given translation domain.

    Given a translation domain, install a _() function using gettext's
    install() function.

    The main difference from gettext.install() is that we allow
    overriding the default localedir (e.g. /usr/share/locale) using
    a translation-domain-specific environment variable (e.g.
    NOVA_LOCALEDIR).
    """
    gettext.install(domain,
                    localedir=os.environ.get(domain.upper() + '_LOCALEDIR'),
                    unicode=True)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from savanna.storage.db import DB


class NodeTemplate(DB.Model):
    __tablename__ = 'NodeTemplate'

    id = DB.Column(DB.String(36), primary_key=True)
    name = DB.Column(DB.String(80), unique=True, nullable=False)
    node_type_id = DB.Column(DB.String(36), DB.ForeignKey('NodeType.id'),
                             nullable=False)
    flavor_id = DB.Column(DB.String(36), nullable=False)

    node_template_configs = DB.relationship('NodeTemplateConfig',
                                            cascade="all,delete",
                                            backref='node_template')
    cluster_node_counts = DB.relationship('ClusterNodeCount',
                                          cascade="all,delete",
                                          backref='node_template')
    nodes = DB.relationship('Node', cascade="all,delete",
                            backref='node_template')

    def __init__(self, name, node_type_id, flavor_id):
        self.id = uuid.uuid4().hex
        self.name = name
        self.node_type_id = node_type_id
        self.flavor_id = flavor_id

    def __repr__(self):
        return '<NodeTemplate %s / %s>' % (self.name, self.node_type_id)


class Cluster(DB.Model):
    __tablename__ = 'Cluster'

    id = DB.Column(DB.String(36), primary_key=True)
    name = DB.Column(DB.String(80), unique=True, nullable=False)
    base_image_id = DB.Column(DB.String(36), nullable=False)
    status = DB.Column(DB.String(80))
    tenant_id = DB.Column(DB.String(36), nullable=False)

    nodes = DB.relationship('Node', cascade="all,delete", backref='cluster')
    service_urls = DB.relationship('ServiceUrl', cascade="all,delete",
                                   backref='cluster')
    node_counts = DB.relationship('ClusterNodeCount', cascade="all,delete",
                                  backref='cluster')

    # node_templates: [(node_template_id, count), ...]

    def __init__(self, name, base_image_id, tenant_id, status=None):
        self.id = uuid.uuid4().hex
        self.name = name
        self.base_image_id = base_image_id
        if not status:
            status = 'Starting'
        self.status = status
        self.tenant_id = tenant_id

    def __repr__(self):
        return '<Cluster %s / %s>' % (self.name, self.status)


NODE_TYPE_NODE_PROCESS = DB.Table('NodeType_NodeProcess', DB.metadata,
                                  DB.Column('node_type_id', DB.String(36),
                                            DB.ForeignKey('NodeType.id')),
                                  DB.Column('node_process_id', DB.String(36),
                                            DB.ForeignKey('NodeProcess.id')))


class NodeType(DB.Model):
    __tablename__ = 'NodeType'

    id = DB.Column(DB.String(36), primary_key=True)
    name = DB.Column(DB.String(80), unique=True, nullable=False)
    processes = DB.relationship('NodeProcess',
                                cascade="all,delete",
                                secondary=NODE_TYPE_NODE_PROCESS,
                                backref='node_types')
    node_templates = DB.relationship('NodeTemplate', cascade="all,delete",
                                     backref='node_type')

    def __init__(self, name):
        self.id = uuid.uuid4().hex
        self.name = name

    def __repr__(self):
        return '<NodeType %s>' % self.name


class NodeProcess(DB.Model):
    __tablename__ = 'NodeProcess'

    id = DB.Column(DB.String(36), primary_key=True)
    name = DB.Column(DB.String(80), unique=True, nullable=False)
    node_process_properties = DB.relationship('NodeProcessProperty',
                                              cascade="all,delete",
                                              backref='node_process')

    def __init__(self, name):
        self.id = uuid.uuid4().hex
        self.name = name

    def __repr__(self):
        return '<NodeProcess %s>' % self.name


class NodeProcessProperty(DB.Model):
    __tablename__ = 'NodeProcessProperty'
    __table_args__ = (
        DB.UniqueConstraint('node_process_id', 'name'),
    )

    id = DB.Column(DB.String(36), primary_key=True)
    node_process_id = DB.Column(DB.String(36), DB.ForeignKey('NodeProcess.id'))
    name = DB.Column(DB.String(80), nullable=False)
    required = DB.Column(DB.Boolean, nullable=False)
    default = DB.Column(DB.String(36))
    node_template_configs = DB.relationship('NodeTemplateConfig',
                                            cascade="all,delete",
                                            backref='node_process_property')

    def __init__(self, node_process_id, name, required=True, default=None):
        self.id = uuid.uuid4().hex
        self.node_process_id = node_process_id
        self.name = name
        self.required = required
        self.default = default

    def __repr__(self):
        return '<NodeProcessProperty %s>' % self.name


class NodeTemplateConfig(DB.Model):
    __tablename__ = 'NodeTemplateConfig'
    __table_args__ = (
        DB.UniqueConstraint('node_template_id', 'node_process_property_id'),
    )

    id = DB.Column(DB.String(36), primary_key=True)
    node_template_id = DB.Column(
        DB.String(36),
        DB.ForeignKey('NodeTemplate.id'))
    node_process_property_id = DB.Column(
        DB.String(36),
        DB.ForeignKey('NodeProcessProperty.id'))
    value = DB.Column(DB.String(36))

    def __init__(self, node_template_id, node_process_property_id, value):
        self.id = uuid.uuid4().hex
        self.node_template_id = node_template_id
        self.node_process_property_id = node_process_property_id
        self.value = value

    def __repr__(self):
        return '<NodeTemplateConfig %s.%s / %s>' \
               % (self.node_template_id, self.node_process_property_id,
                  self.value)


class ClusterNodeCount(DB.Model):
    __tablename__ = 'ClusterNodeCount'
    __table_args__ = (
        DB.UniqueConstraint('cluster_id', 'node_template_id'),
    )

    id = DB.Column(DB.String(36), primary_key=True)
    cluster_id = DB.Column(DB.String(36), DB.ForeignKey('Cluster.id'))
    node_template_id = DB.Column(DB.String(36),
                                 DB.ForeignKey('NodeTemplate.id'))
    count = DB.Column(DB.Integer, nullable=False)

    def __init__(self, cluster_id, node_template_id, count):
        self.id = uuid.uuid4().hex
        self.cluster_id = cluster_id
        self.node_template_id = node_template_id
        self.count = count

    def __repr__(self):
        return '<ClusterNodeCount %s / %s>' \
               % (self.node_template_id, self.count)


class Node(DB.Model):
    __tablename__ = 'Node'

    # do we need own id?
    vm_id = DB.Column(DB.String(36), primary_key=True)
    cluster_id = DB.Column(DB.String(36), DB.ForeignKey('Cluster.id'))
    node_template_id = DB.Column(DB.String(36),
                                 DB.ForeignKey('NodeTemplate.id'))

    def __init__(self, vm_id, cluster_id, node_template_id):
        self.vm_id = vm_id
        self.cluster_id = cluster_id
        self.node_template_id = node_template_id

    def __repr__(self):
        return '<Node based on %s>' % self.node_template.name


class ServiceUrl(DB.Model):
    __tablename__ = 'ServiceUrl'

    id = DB.Column(DB.String(36), primary_key=True)
    cluster_id = DB.Column(DB.String(36), DB.ForeignKey('Cluster.id'))
    name = DB.Column(DB.String(80))
    url = DB.Column(DB.String(80), nullable=False)

    def __init__(self, cluster_id, name, url):
        self.id = uuid.uuid4().hex
        self.cluster_id = cluster_id
        self.name = name
        self.url = url

    def __repr__(self):
        return '<ServiceUrl %s / %s>' % (self.name, self.url)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import Flask

from oslo.config import cfg
from savanna.openstack.common import log
from savanna.storage.db import DB
from savanna.storage.db import setup_storage
from savanna.storage.defaults import setup_defaults

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class BaseCmd(object):
    name = None

    @classmethod
    def add_argument_parser(cls, subparsers):
        parser = subparsers.add_parser(cls.name, help=cls.__doc__)
        parser.set_defaults(cmd_class=cls)
        return parser


class ResetDbCmd(BaseCmd):
    """Reset the database."""

    name = 'reset-db'

    @classmethod
    def add_argument_parser(cls, subparsers):
        parser = super(ResetDbCmd, cls).add_argument_parser(subparsers)
        parser.add_argument('--with-gen-templates', action='store_true')
        return parser

    @staticmethod
    def main():
        gen = CONF.command.with_gen_templates

        app = Flask('savanna.manage')
        setup_storage(app)

        DB.drop_all()
        DB.create_all()

        setup_defaults(True, gen)

        LOG.info("DB has been removed and created from scratch, "
                 "gen templates: %s", gen)


CLI_COMMANDS = [
    ResetDbCmd,
]


def add_command_parsers(subparsers):
    for cmd in CLI_COMMANDS:
        cmd.add_argument_parser(subparsers)


command_opt = cfg.SubCommandOpt('command',
                                title='Commands',
                                help='Available commands',
                                handler=add_command_parsers)


def main(argv=None, config_files=None):
    CONF.register_cli_opt(command_opt)
    CONF(args=argv[1:],
         project='savanna',
         usage='%(prog)s [' + '|'.join(
             [cmd.name for cmd in CLI_COMMANDS]) + ']',
         default_config_files=config_files)
    log.setup("savanna")
    CONF.command.cmd_class.main()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
import json
from keystoneclient.v2_0 import Client as keystone_client
import requests
import savanna.tests.integration.parameters as param
import unittest


class ITestCase(unittest.TestCase):

    def setUp(self):
        self.port = param.SAVANNA_PORT
        self.host = param.SAVANNA_HOST

        self.maxDiff = None

        self.baseurl = 'http://' + self.host + ':' + self.port

        self.keystone = keystone_client(
            username=param.OS_USERNAME,
            password=param.OS_PASSWORD,
            tenant_name=param.OS_TENANT_NAME,
            auth_url=param.OS_AUTH_URL
        )

        self.tenant = self.keystone.tenant_id
        self.token = self.keystone.auth_token

        self.flavor_id = param.FLAVOR_ID
        self.image_id = param.IMAGE_ID

        self.url_nt = '/v0.2/%s/node-templates' % self.tenant
        self.url_nt_with_slash = '/v0.2/%s/node-templates/' % self.tenant
        self.url_cluster = '/v0.2/%s/clusters' % self.tenant
        self.url_cl_with_slash = '/v0.2/%s/clusters/' % self.tenant

#----------------------CRUD_comands--------------------------------------------

    def post(self, url, body):
        URL = self.baseurl + url
        resp = requests.post(URL, data=body, headers={
            'x-auth-token': self.token, 'Content-Type': 'application/json'})
        data = json.loads(resp.content) if resp.status_code == 202 \
            else resp.content
        print('URL = %s\ndata = %s\nresponse = %s\ndata = %s\n'
              % (URL, body, resp.status_code, data))
        return resp

    def put(self, url, body):
        URL = self.baseurl + url
        resp = requests.put(URL, data=body, headers={
            'x-auth-token': self.token, 'Content-Type': 'application/json'})
        data = json.loads(resp.content)
        print('URL = %s\ndata = %s\nresponse = %s\ndata = %s\n'
              % (URL, body, resp.status_code, data))
        return resp

    def get(self, url, printing):
        URL = self.baseurl + url
        resp = requests.get(URL, headers={'x-auth-token': self.token})
        if printing:
            print('URL = %s\nresponse = %s\n' % (URL, resp.status_code))
        if resp.status_code != 200:
            data = json.loads(resp.content)
            print('data= %s\n') % data
        return resp

    def delete(self, url):
        URL = self.baseurl + url
        resp = requests.delete(URL, headers={'x-auth-token': self.token})
        print('URL = %s\nresponse = %s\n' % (URL, resp.status_code))
        if resp.status_code != 204:
            data = json.loads(resp.content)
            print('data= %s\n') % data
        return resp

    def _post_object(self, url, body, code):
        post = self.post(url, json.dumps(body))
        self.assertEquals(post.status_code, code)
        data = json.loads(post.content)
        return data

    def _get_object(self, url, obj_id, code, printing=False):
        rv = self.get(url + obj_id, printing)
        self.assertEquals(rv.status_code, code)
        data = json.loads(rv.content)
        return data

    def _del_object(self, url, obj_id, code):
        rv = self.delete(url + obj_id)
        self.assertEquals(rv.status_code, code)
        if rv.status_code != 204:
            data = json.loads(rv.content)
            return data
        else:
            code = self.delete(url + obj_id).status_code
            while code != 404:
                eventlet.sleep(1)
                code = self.delete(url + obj_id).status_code

#----------------------other_commands------------------------------------------

    def _get_body_nt(self, name, nt_type, hs1, hs2):
        node = 'name' if nt_type in ['JT+NN', 'NN'] else 'data'
        tracker = 'job' if nt_type in ['JT+NN', 'JT'] else 'task'
        processes_name = nt_type
        nt = {
            u'name': u'%s.%s' % (name, param.FLAVOR_ID),
            u'%s_node' % node: {u'heap_size': u'%d' % hs1},
            u'%s_tracker' % tracker: {u'heap_size': u'%d' % hs2},
            u'node_type': {
                u'processes': [u'%s_tracker' % tracker,
                               u'%s_node' % node],
                u'name': u'%s' % processes_name},
            u'flavor_id': u'%s' % self.flavor_id
        }
        if nt_type == 'NN':
            del nt[u'%s_tracker' % tracker]
            nt[u'node_type'][u'processes'] = [u'%s_node' % node]
        elif nt_type == 'JT':
            del nt[u'%s_node' % node]
            nt[u'node_type'][u'processes'] = [u'%s_tracker' % tracker]
        return nt

    def _get_body_cluster(self, name, master_name, worker_name, node_number):
        return {
            u'status': u'Starting',
            u'service_urls': {},
            u'name': u'%s' % name,
            u'base_image_id': u'%s' % self.image_id,
            u'node_templates':
            {
                u'%s.%s' % (master_name, param.FLAVOR_ID): 1,
                u'%s.%s' % (worker_name, param.FLAVOR_ID): node_number
            },
            u'nodes': []
        }

    def change_field_nt(self, data, old_field, new_field):
        val = data['node_template'][old_field]
        del data['node_template'][old_field]
        data['node_template'][new_field] = val
        return data

    def make_nt(self, nt_name, node_type, jt_heap_size, nn_heap_size):
        nt = dict(
            node_template=dict(
                name='%s.%s' % (nt_name, param.FLAVOR_ID),
                node_type='JT+NN',
                flavor_id=self.flavor_id,
                job_tracker={
                    'heap_size': '%d' % jt_heap_size
                },
                name_node={
                    'heap_size': '%d' % nn_heap_size
                }
            ))
        if node_type == 'TT+DN':
            nt['node_template']['node_type'] = 'TT+DN'
            nt = self.change_field_nt(nt, 'job_tracker', 'task_tracker')
            nt = self.change_field_nt(nt, 'name_node', 'data_node')
        elif node_type == 'NN':
            nt['node_template']['node_type'] = 'NN'
            del nt['node_template']['job_tracker']
        elif node_type == 'JT':
            nt['node_template']['node_type'] = 'JT'
            del nt['node_template']['name_node']
        return nt

    def make_cluster_body(self, cluster_name, name_master_node,
                          name_worker_node, number_workers):
        body = dict(
            cluster=dict(
                name=cluster_name,
                base_image_id=self.image_id,
                node_templates={
                    '%s.%s' % (name_master_node, param.FLAVOR_ID): 1,
                    '%s.%s' %
                    (name_worker_node, param.FLAVOR_ID): number_workers
                }
            ))
        return body

    def delete_node_template(self, data):
        data = data['node_template']
        object_id = data.pop(u'id')
        self._del_object(self.url_nt_with_slash, object_id, 204)

    def _crud_object(self, body, get_body, url):
        data = self._post_object(url, body, 202)
        get_url = None
        object_id = None
        try:
            obj = 'node_template' if url == self.url_nt else 'cluster'
            get_url = self.url_nt_with_slash if url == self.url_nt \
                else self.url_cl_with_slash
            data = data['%s' % obj]
            object_id = data.pop(u'id')
            self.assertEquals(data, get_body)
            get_data = self._get_object(get_url, object_id, 200)
            get_data = get_data['%s' % obj]
            del get_data[u'id']
            if obj == 'cluster':
                self._await_cluster_active(
                    get_body, get_data, get_url, object_id)
        except Exception as e:
            self.fail('failure:' + str(e))
        finally:
            self._del_object(get_url, object_id, 204)
        return object_id

    def _await_cluster_active(self, get_body, get_data, get_url, object_id):
        get_body[u'status'] = u'Active'
        del get_body[u'service_urls']
        del get_body[u'nodes']
        i = 1
        while get_data[u'status'] != u'Active':
            if i > int(param.TIMEOUT) * 6:
                self.fail(
                    'cluster not Starting -> Active, passed %d minutes'
                    % param.TIMEOUT)
            get_data = self._get_object(get_url, object_id, 200)
            get_data = get_data['cluster']
            del get_data[u'id']
            del get_data[u'service_urls']
            del get_data[u'nodes']
            eventlet.sleep(10)
            i += 1
        self.assertEquals(get_data, get_body)



from keystoneclient.v2_0 import Client as keystone_client
import os
from oslo.config import cfg
import sys


possible_topdir = os.path.normpath(os.path.join(os.path.abspath(sys.argv[0]),
                                                os.pardir,
                                                os.pardir))
if os.path.exists(os.path.join(possible_topdir,
                               'savanna',
                               '__init__.py')):
    sys.path.insert(0, possible_topdir)

cli_opts = [
    cfg.StrOpt('username', default='',
               help='set username'),
    cfg.StrOpt('password', default='',
               help='set password'),
    cfg.StrOpt('tenant', default='',
               help='set tenant'),
]

CONF = cfg.CONF
CONF.import_opt('os_admin_username', 'savanna.main')
CONF.import_opt('os_admin_password', 'savanna.main')
CONF.import_opt('os_admin_tenant_name', 'savanna.main')
CONF.register_cli_opts(cli_opts)


def main():
    dev_conf = os.path.join(possible_topdir,
                            'etc',
                            'savanna',
                            'savanna.conf')
    config_files = None
    if os.path.exists(dev_conf):
        config_files = [dev_conf]

    CONF(sys.argv[1:], project='get_auth_token',
         default_config_files=config_files)

    user = CONF.username or CONF.os_admin_username
    password = CONF.password or CONF.os_admin_password
    tenant = CONF.tenant or CONF.os_admin_tenant_name

    protocol = CONF.os_auth_protocol
    host = CONF.os_auth_host
    port = CONF.os_auth_port

    auth_url = "%s://%s:%s/v2.0/" % (protocol, host, port)

    print "User: %s" % user
    print "Password: %s" % password
    print "Tenant: %s" % tenant
    print "Auth URL: %s" % auth_url

    keystone = keystone_client(
        username=user,
        password=password,
        tenant_name=tenant,
        auth_url=auth_url
    )

    result = keystone.authenticate()

    print "Auth succeed: %s" % result
    print "Auth token: %s" % keystone.auth_token
    print "Tenant [%s] id: %s" % (tenant, keystone.tenant_id)


if __name__ == "__main__":
    main()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna.storage.db import DB

from savanna.storage.models import NodeTemplate, NodeProcess, Cluster, \
    ClusterNodeCount, NodeTemplateConfig, NodeType, NodeProcessProperty


## Node Template ops:

def get_node_template(**args):
    return NodeTemplate.query.filter_by(**args).first()


def get_node_templates(**args):
    return NodeTemplate.query.filter_by(**args).all()


def is_node_template_associated(**args):
    nt = get_node_template(**args)
    return nt and (len(nt.nodes) or len(nt.cluster_node_counts))


def create_node_template(name, node_type_id, flavor_id, configs):
    """Creates new node templates.

    :param name: template name
    :param node_type_id: node type
    :param flavor_id: flavor
    :param configs: dict of process->property->value
    :return: created node template
    """
    node_template = NodeTemplate(name, node_type_id, flavor_id)
    DB.session.add(node_template)
    for process_name in configs:
        process = NodeProcess.query.filter_by(name=process_name).first()
        conf = configs.get(process_name)
        for prop in process.node_process_properties:
            val = conf.get(prop.name, None)
            if not val and prop.required:
                if not prop.default:
                    raise RuntimeError('Template \'%s\', value missed '
                                       'for required param: %s %s'
                                       % (name, process.name, prop.name))
                val = prop.default
            DB.session.add(NodeTemplateConfig(node_template.id, prop.id, val))
    DB.session.commit()

    return node_template


def terminate_node_template(**args):
    template = get_node_template(**args)
    if template:
        DB.session.delete(template)
        DB.session.commit()
        return True
    else:
        return False


## Cluster ops:

def get_cluster(**args):
    return Cluster.query.filter_by(**args).first()


def get_clusters(**args):
    return Cluster.query.filter_by(**args).all()


def create_cluster(name, base_image_id, tenant_id, templates):
    """Creates new cluster.

    :param name: cluster name
    :param base_image_id: base image
    :param tenant_id: tenant
    :param templates: dict of template->count
    :return: created cluster
    """
    cluster = Cluster(name, base_image_id, tenant_id)
    DB.session.add(cluster)
    for template in templates:
        count = templates.get(template)
        template_id = get_node_template(name=template).id
        cnc = ClusterNodeCount(cluster.id, template_id, int(count))
        DB.session.add(cnc)
    DB.session.commit()

    return cluster


def terminate_cluster(**args):
    cluster = get_cluster(**args)
    DB.session.delete(cluster)
    DB.session.commit()


def update_cluster_status(new_status, **args):
    cluster = Cluster.query.filter_by(**args).first()
    cluster.status = new_status
    DB.session.add(cluster)
    DB.session.commit()

    return cluster


## Node Process ops:

def create_node_process(name, properties):
    """Creates new node process and node process properties.

    :param name: process name
    :param properties: array of triples (name, required, default)
    :return: created node process
    """
    process = NodeProcess(name)
    DB.session.add(process)
    DB.session.commit()
    for p in properties:
        prop = NodeProcessProperty(process.id, p[0], p[1], p[2])
        DB.session.add(prop)
    DB.session.commit()

    return process


## Node Type ops:

def get_node_type(**args):
    return NodeType.query.filter_by(**args).first()


def get_node_types(**args):
    return NodeType.query.filter_by(**args).all()


def create_node_type(name, processes):
    """Creates new node type using specified list of processes

    :param name:
    :param processes:
    :return:
    """
    node_type = NodeType(name)
    node_type.processes = processes
    DB.session.add(node_type)
    DB.session.commit()

    return node_type



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna.storage.storage import create_node_type, \
    create_node_template, create_node_process
from savanna.openstack.common import log as logging

LOG = logging.getLogger(__name__)


def setup_defaults(reset_db=False, gen_templates=False):
    nt_jt_nn = None
    nt_jt = None
    nt_nn = None
    nt_tt_dn = None

    if reset_db:
        # setup default processes
        p_jt = create_node_process('job_tracker',
                                   [('heap_size', True, None),
                                    ('mapred.job.tracker.handler.count',
                                     False, None)])
        p_nn = create_node_process('name_node',
                                   [('heap_size', True, None),
                                    ('dfs.namenode.handler.count',
                                     False, None),
                                    ('dfs.block.size', False, None),
                                    ('dfs.replication', False, None)])
        p_tt = create_node_process('task_tracker',
                                   [('heap_size', True, None),
                                    ('mapred.child.java.opts', False, None),
                                    ('mapred.map.tasks', False, None),
                                    ('mapred.tasktracker.map.tasks.maximum',
                                     False, None),
                                    ('mapred.reduce.tasks', False, None),
                                    ('mapred.tasktracker.reduce.tasks.maximum',
                                     False, None)])

        p_dn = create_node_process('data_node',
                                   [('heap_size', True, None),
                                    ('dfs.datanode.max.xcievers', False, None),
                                    ('dfs.block.size', False, None),
                                    ('dfs.replication', False, None),
                                    ('dfs.datanode.handler.count',
                                     False, None)])

        for p in [p_jt, p_nn, p_tt, p_dn]:
            LOG.info('New NodeProcess: \'%s\'', p.name)

        # setup default node types
        nt_jt_nn = create_node_type('JT+NN', [p_jt, p_nn])
        nt_jt = create_node_type('JT', [p_jt])
        nt_nn = create_node_type('NN', [p_nn])
        nt_tt_dn = create_node_type('TT+DN', [p_tt, p_dn])

        for nt in [nt_jt_nn, nt_jt, nt_nn, nt_tt_dn]:
            LOG.info('New NodeType: \'%s\' %s',
                     nt.name, [p.name.__str__() for p in nt.processes])

    if gen_templates:
        _generate_templates(nt_jt_nn, nt_jt, nt_nn, nt_tt_dn)

    LOG.info('All defaults has been inserted')


def _generate_templates(nt_jt_nn, nt_jt, nt_nn, nt_tt_dn):
    jt_nn_small = create_node_template('jt_nn.small', nt_jt_nn.id, 'm1.small',
                                       {
                                           'job_tracker': {
                                               'heap_size': '896'
                                           },
                                           'name_node': {
                                               'heap_size': '896'
                                           }
                                       })
    jt_nn_medium = create_node_template('jt_nn.medium', nt_jt_nn.id,
                                        'm1.medium',
                                        {
                                            'job_tracker': {
                                                'heap_size': '1792'
                                            },
                                            'name_node': {
                                                'heap_size': '1792'
                                            }
                                        })
    tt_dn_small = create_node_template('tt_dn.small', nt_tt_dn.id, 'm1.small',
                                       {
                                           'task_tracker': {
                                               'heap_size': '896'
                                           },
                                           'data_node': {
                                               'heap_size': '896'
                                           }
                                       })
    tt_dn_medium = create_node_template('tt_dn.medium', nt_tt_dn.id,
                                        'm1.medium',
                                        {
                                            'task_tracker': {
                                                'heap_size': '1792'
                                            },
                                            'data_node': {
                                                'heap_size': '1792'
                                            }
                                        })

    for tmpl in [jt_nn_small, jt_nn_medium, tt_dn_small, tt_dn_medium]:
        LOG.info('New NodeTemplate: \'%s\' %s', tmpl.name, tmpl.flavor_id)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
import unittest
import uuid

import eventlet
from oslo.config import cfg

import savanna.main
from savanna.openstack.common import log as logging
from savanna.service import api
from savanna.storage.db import DB
from savanna.storage.defaults import setup_defaults
from savanna.storage.models import Node, NodeTemplate
from savanna.utils.openstack import nova
from savanna.utils import scheduler

LOG = logging.getLogger(__name__)


def _stub_vm_creation_job(template_id):
    template = NodeTemplate.query.filter_by(id=template_id).first()
    eventlet.sleep(2)
    return 'ip-address', uuid.uuid4().hex, template.id


def _stub_launch_cluster(headers, cluster):
    LOG.debug('stub launch_cluster called with %s, %s', headers, cluster)
    pile = eventlet.GreenPile(scheduler.POOL)

    for elem in cluster.node_counts:
        node_count = elem.count
        for _ in xrange(0, node_count):
            pile.spawn(_stub_vm_creation_job, elem.node_template_id)

    for (ip, vm_id, elem) in pile:
        DB.session.add(Node(vm_id, cluster.id, elem))
        LOG.debug("VM '%s/%s/%s' created", ip, vm_id, elem)

    return True


def _stub_stop_cluster(headers, cluster):
    LOG.debug("stub stop_cluster called with %s, %s", headers, cluster)


def _stub_auth_token(*args, **kwargs):
    LOG.debug('stub token filter called with %s, %s', args, kwargs)

    def _filter(app):
        def _handler(env, start_response):
            env['HTTP_X_TENANT_ID'] = 'tenant-id-1'
            return app(env, start_response)

        return _handler

    return _filter


def _stub_auth_valid(*args, **kwargs):
    LOG.debug('stub token validation called with %s, %s', args, kwargs)

    def _filter(app):
        def _handler(env, start_response):
            return app(env, start_response)

        return _handler

    return _filter


def _stub_get_flavors(headers):
    LOG.debug('Stub get_flavors called with %s', headers)
    return [u'test_flavor', u'test_flavor_2']


def _stub_get_images(headers):
    LOG.debug('Stub get_images called with %s', headers)
    return [u'base-image-id', u'base-image-id_2']


def _stub_get_limits(headers):
    limits = dict(maxTotalCores=100,
                  maxTotalRAMSize=51200,
                  maxTotalInstances=100,
                  totalCoresUsed=0,
                  totalRAMUsed=0,
                  totalInstancesUsed=0)

    LOG.debug('Stub get_limits called with headers %s and limits %s',
              headers, limits)

    return limits


class StubFlavor:
    def __init__(self, name, vcpus, ram):
        self.name = name
        self.vcpus = int(vcpus)
        self.ram = int(ram)


def _stub_get_flavor(headers, **kwargs):
    return StubFlavor('test_flavor', 1, 512)


CONF = cfg.CONF
CONF.import_opt('debug', 'savanna.openstack.common.log')
CONF.import_opt('allow_cluster_ops', 'savanna.config')
CONF.import_opt('database_uri', 'savanna.storage.db', group='sqlalchemy')
CONF.import_opt('echo', 'savanna.storage.db', group='sqlalchemy')


class SavannaTestCase(unittest.TestCase):
    def setUp(self):
        self.db_fd, self.db_path = tempfile.mkstemp()
        self.maxDiff = 10000

        # override configs
        CONF.set_override('debug', True)
        CONF.set_override('allow_cluster_ops', True)  # stub process
        CONF.set_override('database_uri', 'sqlite:///' + self.db_path,
                          group='sqlalchemy')
        CONF.set_override('echo', False, group='sqlalchemy')

        # store functions that will be stubbed
        self._prev_auth_token = savanna.main.auth_token
        self._prev_auth_valid = savanna.main.auth_valid
        self._prev_cluster_launch = api.cluster_ops.launch_cluster
        self._prev_cluster_stop = api.cluster_ops.stop_cluster
        self._prev_get_flavors = nova.get_flavors
        self._prev_get_images = nova.get_images
        self._prev_get_limits = nova.get_limits
        self._prev_get_flavor = nova.get_flavor

        # stub functions
        savanna.main.auth_token = _stub_auth_token
        savanna.main.auth_valid = _stub_auth_valid
        api.cluster_ops.launch_cluster = _stub_launch_cluster
        api.cluster_ops.stop_cluster = _stub_stop_cluster
        nova.get_flavors = _stub_get_flavors
        nova.get_images = _stub_get_images
        nova.get_limits = _stub_get_limits
        nova.get_flavor = _stub_get_flavor

        app = savanna.main.make_app()

        DB.drop_all()
        DB.create_all()
        setup_defaults(True, True)

        LOG.debug('Test db path: %s', self.db_path)
        LOG.debug('Test app.config: %s', app.config)

        self.app = app.test_client()

    def tearDown(self):
        # unstub functions
        savanna.main.auth_token = self._prev_auth_token
        savanna.main.auth_valid = self._prev_auth_valid
        api.cluster_ops.launch_cluster = self._prev_cluster_launch
        api.cluster_ops.stop_cluster = self._prev_cluster_stop
        nova.get_flavors = self._prev_get_flavors
        nova.get_images = self._prev_get_images
        nova.get_limits = self._prev_get_limits
        nova.get_flavor = self._prev_get_flavor

        os.close(self.db_fd)
        os.unlink(self.db_path)

        # place back default configs
        CONF.clear_override('debug')
        CONF.clear_override('allow_cluster_ops')
        CONF.clear_override('database_uri', group='sqlalchemy')
        CONF.clear_override('echo', group='sqlalchemy')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from eventlet import monkey_patch
from flask import Flask
from keystoneclient.middleware.auth_token import filter_factory as auth_token
from oslo.config import cfg
from werkzeug.exceptions import default_exceptions
from werkzeug.exceptions import HTTPException

from savanna.api import v02 as api_v02
from savanna.middleware.auth_valid import filter_factory as auth_valid
from savanna.storage.db import setup_storage
from savanna.utils.api import render
from savanna.utils.scheduler import setup_scheduler

from savanna.openstack.common import log

LOG = log.getLogger(__name__)

monkey_patch(os=True, select=True, socket=True, thread=True, time=True)

opts = [
    cfg.StrOpt('os_auth_protocol',
               default='http',
               help='Protocol used to access OpenStack Identity service'),
    cfg.StrOpt('os_auth_host',
               default='openstack',
               help='IP or hostname of machine on which OpenStack Identity '
                    'service is located'),
    cfg.StrOpt('os_auth_port',
               default='35357',
               help='Port of OpenStack Identity service'),
    cfg.StrOpt('os_admin_username',
               default='admin',
               help='This OpenStack user is used to verify provided tokens. '
                    'The user must have admin role in <os_admin_tenant_name> '
                    'tenant'),
    cfg.StrOpt('os_admin_password',
               default='nova',
               help='Password of the admin user'),
    cfg.StrOpt('os_admin_tenant_name',
               default='admin',
               help='Name of tenant where the user is admin')
]

CONF = cfg.CONF
CONF.register_opts(opts)


def make_app():
    """App builder (wsgi)

    Entry point for Savanna REST API server
    """
    app = Flask('savanna.api')

    @app.route('/', methods=['GET'])
    def version_list():
        return render({
            "versions": [
                {"id": "v0.2", "status": "CURRENT"}
            ]
        })

    app.register_blueprint(api_v02.rest, url_prefix='/v0.2')

    setup_storage(app)
    setup_scheduler(app)

    def make_json_error(ex):
        status_code = (ex.code
                       if isinstance(ex, HTTPException)
                       else 500)
        description = (ex.description
                       if isinstance(ex, HTTPException)
                       else str(ex))
        return render({'error': status_code, 'error_message': description},
                      status=status_code)

    for code in default_exceptions.iterkeys():
        app.error_handler_spec[None][code] = make_json_error

    app.wsgi_app = auth_valid(app.config)(app.wsgi_app)

    app.wsgi_app = auth_token(
        app.config,
        auth_host=CONF.os_auth_host,
        auth_port=CONF.os_auth_port,
        auth_protocol=CONF.os_auth_protocol,
        admin_user=CONF.os_admin_username,
        admin_password=CONF.os_admin_password,
        admin_tenant=CONF.os_admin_tenant_name
    )(app.wsgi_app)

    return app



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask.ext.sqlalchemy import SQLAlchemy
from oslo.config import cfg

DB = SQLAlchemy()

opts = [
    cfg.StrOpt('database_uri',
               default='sqlite:////tmp/savanna.db',
               help='URL for sqlalchemy database'),
    cfg.BoolOpt('echo',
                default=False,
                help='Sqlalchemy echo')
]

CONF = cfg.CONF
CONF.register_opts(opts, group='sqlalchemy')


def setup_storage(app):
    app.config['SQLALCHEMY_DATABASE_URI'] = CONF.sqlalchemy.database_uri
    app.config['SQLALCHEMY_ECHO'] = CONF.sqlalchemy.echo

    DB.app = app
    DB.init_app(app)
    DB.create_all()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna.tests.integration.db import ITestCase
from telnetlib import Telnet


class ITestNodeTemplateApi(ITestCase):

    def setUp(self):
        super(ITestNodeTemplateApi, self).setUp()
        Telnet(self.host, self.port)

    def test_crud_nt_jtnn(self):
        nt_jtnn = self.make_nt('jtnn', 'JT+NN', 1024, 1024)
        get_jtnn = self._get_body_nt('jtnn', 'JT+NN', 1024, 1024)

        self._crud_object(nt_jtnn, get_jtnn, self.url_nt)

    def test_crud_nt_ttdn(self):
        nt_ttdn = self.make_nt('ttdn', 'TT+DN', 1024, 1024)
        get_ttdn = self._get_body_nt('ttdn', 'TT+DN', 1024, 1024)

        self._crud_object(nt_ttdn, get_ttdn, self.url_nt)

    def test_crud_nt_nn(self):
        nt_nn = self.make_nt('nn', 'NN', 1024, 1024)
        get_nn = self._get_body_nt('nn', 'NN', 1024, 1024)

        self._crud_object(nt_nn, get_nn, self.url_nt)

    def test_crud_nt_jt(self):
        nt_jt = self.make_nt('jt', 'JT', 1024, 1024)
        get_jt = self._get_body_nt('jt', 'JT', 1024, 1024)

        self._crud_object(nt_jt, get_jt, self.url_nt)



import savanna.openstack.common.importutils as importutils

_CONF = importutils.try_import('savanna.tests.integration.config')


def _get_conf(key, default):
    return getattr(_CONF, key) if _CONF and hasattr(_CONF, key) else default

OS_USERNAME = _get_conf('OS_USERNAME', 'admin')
OS_PASSWORD = _get_conf('OS_PASSWORD', 'password')
OS_TENANT_NAME = _get_conf('OS_TENANT_NAME', 'admin')
OS_AUTH_URL = _get_conf('OS_AUTH_URL', 'http://localhost:35357/v2.0/')

SAVANNA_HOST = _get_conf('SAVANNA_HOST', '192.168.1.1')
SAVANNA_PORT = _get_conf('SAVANNA_PORT', '8080')

IMAGE_ID = _get_conf('IMAGE_ID', '42')
FLAVOR_ID = _get_conf('FLAVOR_ID', 'abc')

NODE_USERNAME = _get_conf('NODE_USERNAME', 'username')
NODE_PASSWORD = _get_conf('NODE_PASSWORD', 'password')

CLUSTER_NAME_CRUD = _get_conf('CLUSTER_NAME_CRUD', 'cluster-crud')
CLUSTER_NAME_HADOOP = _get_conf('CLUSTER_NAME_HADOOP', 'cluster-hadoop')

IP_PREFIX = _get_conf('IP_PREFIX', '10.')

TIMEOUT = _get_conf('TIMEOUT', '15')

HADOOP_VERSION = _get_conf('HADOOP_VERSION', '1.1.1')
HADOOP_DIRECTORY = _get_conf('HADOOP_DIRECTORY', '/usr/share/hadoop')
HADOOP_LOG_DIRECTORY = _get_conf('HADOOP_LOG_DIRECTORY',
                                 '/mnt/log/hadoop/hadoop/userlogs')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo.config import cfg

cli_opts = [
    cfg.StrOpt('host', default='',
               help='set host'),
    cfg.IntOpt('port', default=8080,
               help='set port'),
    cfg.BoolOpt('allow-cluster-ops', default=True,
                help='without that option'
                     ' the application operates in dry run mode and does not '
                     ' send any requests to the OpenStack cluster')
]

CONF = cfg.CONF
CONF.register_cli_opts(cli_opts)


def parse_args(argv, conf_files):
    CONF(argv, project='savanna', default_config_files=conf_files)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def patch_minidom_writexml():
    """Patch for xml.dom.minidom toprettyxml bug with whitespaces around text

    (This patch will be applied for all Python versions < 2.7.3)

    Issue: http://bugs.python.org/issue4147
    Patch: http://hg.python.org/cpython/rev/cb6614e3438b/
    Description: http://ronrothman.com/public/leftbraned/xml-dom-minidom-\
                        toprettyxml-and-silly-whitespace/#best-solution
    """

    import sys
    if sys.version_info >= (2, 7, 3):
        return

    from xml.dom.minidom import Element, Node, Text, _write_data

    def writexml(self, writer, indent="", addindent="", newl=""):
        # indent = current indentation
        # addindent = indentation to add to higher levels
        # newl = newline string
        writer.write(indent + "<" + self.tagName)

        attrs = self._get_attributes()
        a_names = attrs.keys()
        a_names.sort()

        for a_name in a_names:
            writer.write(" %s=\"" % a_name)
            _write_data(writer, attrs[a_name].value)
            writer.write("\"")
        if self.childNodes:
            writer.write(">")
            if (len(self.childNodes) == 1
                    and self.childNodes[0].nodeType == Node.TEXT_NODE):
                self.childNodes[0].writexml(writer, '', '', '')
            else:
                writer.write(newl)
                for node in self.childNodes:
                    node.writexml(writer, indent + addindent, addindent, newl)
                writer.write(indent)
            writer.write("</%s>%s" % (self.tagName, newl))
        else:
            writer.write("/>%s" % (newl))

    Element.writexml = writexml

    def writexml(self, writer, indent="", addindent="", newl=""):
        _write_data(writer, "%s%s%s" % (indent, self.data, newl))

    Text.writexml = writexml



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import savanna.openstack.common.exception as ex


class SavannaException(ex.ApiError):
    """Base Exception for the project

    To correctly use this class, inherit from it and define
    a 'message' and 'code' properties.
    """
    message = "An unknown exception occurred"
    code = "UNKNOWN_EXCEPTION"

    def __str__(self):
        return self.message


class NotFoundException(SavannaException):

    # It could be a various property of object which was not found
    value = None

    def __init__(self, value):
        self.code = "NOT_FOUND"
        self.value = value


## Cluster operations exceptions

class NotEnoughResourcesException(SavannaException):
    def __init__(self, list):
        self.message = "Nova available instances=%s, VCPUs=%s, RAM=%s. " \
                       "Requested instances=%s, VCPUs=%s, RAM=%s" % tuple(list)
        self.code = "NOT_ENOUGH_RESOURCES"


class ClusterNameExistedException(SavannaException):
    def __init__(self, value):
        self.message = "Cluster with name '%s' already exists" % value
        self.code = "CLUSTER_NAME_ALREADY_EXISTS"


class ImageNotFoundException(SavannaException):
    def __init__(self, value):
        self.message = "Cannot find image with id '%s'" % value
        self.code = "IMAGE_NOT_FOUND"


class NotSingleNameNodeException(SavannaException):
    def __init__(self, nn_count):
        self.message = "Hadoop cluster should contain only 1 NameNode. " \
                       "Actual NN count is %s" % nn_count
        self.code = "NOT_SINGLE_NAME_NODE"


class NotSingleJobTrackerException(SavannaException):
    def __init__(self, jt_count):
        self.message = "Hadoop cluster should contain only 1 JobTracker. " \
                       "Actual JT count is %s" % jt_count
        self.code = "NOT_SINGLE_JOB_TRACKER"


class ClusterNotFoundException(NotFoundException):
    def __init__(self, value):
        self.value = value
        self.message = "Cluster '%s' not found" % self.value
        self.code = "CLUSTER_NOT_FOUND"


## NodeTemplates operations exceptions

class NodeTemplateNotFoundException(NotFoundException):
    def __init__(self, value):
        self.value = value
        self.message = "NodeTemplate '%s' not found" % self.value
        self.code = "NODE_TEMPLATE_NOT_FOUND"


class NodeTemplateExistedException(SavannaException):
    def __init__(self, value):
        self.message = "NodeTemplate with name '%s' already exists" % value
        self.code = "NODE_TEMPLATE_ALREADY_EXISTS"


class FlavorNotFoundException(SavannaException):
    def __init__(self, value):
        self.message = "Cannot find flavor with name '%s'" % value
        self.code = "FLAVOR_NOT_FOUND"


class DiscrepancyNodeProcessException(SavannaException):
    def __init__(self, value):
        self.message = "Discrepancies in Node Processes. Required: %s" % value
        self.code = "NODE_PROCESS_DISCREPANCY"


class RequiredParamMissedException(SavannaException):
    def __init__(self, process, param):
        self.message = "Required parameter '%s' of process '%s' should be " \
                       "specified" % (param, process)
        self.code = "REQUIRED_PARAM_MISSED"


class AssociatedNodeTemplateTerminationException(SavannaException):
    def __init__(self, value):
        self.message = ("The are active nodes created using NodeTemplate '%s'"
                        " you trying to terminate") % value
        self.code = "ASSOCIATED_NODE_TEMPLATE_TERMINATION"


class ParamNotAllowedException(SavannaException):
    def __init__(self, param, process):
        self.message = "Parameter '%s' of process '%s' is not " \
                       "allowed to change" % (param, process)
        self.code = "PARAM_IS_NOT_ALLOWED"


## NodeTypes operations exceptions

class NodeTypeNotFoundException(NotFoundException):
    def __init__(self, value):
        self.value = value
        self.message = "NodeType '%s' not found" % self.value
        self.code = "NODE_TYPE_NOT_FOUND"



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
from oslo.config import cfg

from savanna import exceptions as ex
from savanna.openstack.common import log as logging
from savanna.service import cluster_ops
import savanna.storage.storage as storage

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_opt('allow_cluster_ops', 'savanna.config')


## Node Template ops:

def get_node_template(**args):
    return _node_template(storage.get_node_template(**args))


def get_node_templates(**args):
    return [_node_template(tmpl) for tmpl
            in storage.get_node_templates(**args)]


def is_node_template_associated(**args):
    return storage.is_node_template_associated(**args)


def create_node_template(values, headers):
    """Creates new node template from values dict.

    :param values: dict
    :return: created node template resource
    """
    values = values.pop('node_template')

    name = values.pop('name')
    node_type_id = storage.get_node_type(name=values.pop('node_type')).id
    flavor_id = values.pop('flavor_id')

    nt = storage.create_node_template(name, node_type_id, flavor_id, values)

    return get_node_template(id=nt.id)


def terminate_node_template(**args):
    return storage.terminate_node_template(**args)


## Cluster ops:

def get_cluster(**args):
    return _cluster(storage.get_cluster(**args))


def get_clusters(**args):
    return [_cluster(cluster) for cluster in
            storage.get_clusters(**args)]


def create_cluster(values, headers):
    values = values.pop('cluster')

    name = values.pop('name')
    base_image_id = values.pop('base_image_id')
    tenant_id = headers['X-Tenant-Id']
    templates = values.pop('node_templates')

    # todo(slukjanov): check that we can create objects in the specified tenant

    cluster = storage.create_cluster(name, base_image_id, tenant_id, templates)

    eventlet.spawn(_cluster_creation_job, headers, cluster.id)

    return get_cluster(id=cluster.id)


def _cluster_creation_job(headers, cluster_id):
    cluster = storage.get_cluster(id=cluster_id)
    LOG.debug("Starting cluster '%s' creation: %s", cluster_id,
              _cluster(cluster).dict)

    if CONF.allow_cluster_ops:
        launched = cluster_ops.launch_cluster(headers, cluster)
    else:
        LOG.info("Cluster ops are disabled, use --allow-cluster-ops flag")
        launched = True

    if launched:
        storage.update_cluster_status('Active', id=cluster.id)


def terminate_cluster(headers, **args):
    cluster = storage.update_cluster_status('Stopping', **args)

    eventlet.spawn(_cluster_termination_job, headers, cluster.id)


def _cluster_termination_job(headers, cluster_id):
    cluster = storage.get_cluster(id=cluster_id)
    LOG.debug("Stopping cluster '%s' creation: %s", cluster_id,
              _cluster(cluster).dict)

    if CONF.allow_cluster_ops:
        cluster_ops.stop_cluster(headers, cluster)
    else:
        LOG.info("Cluster ops are disabled, use --allow-cluster-ops flag")

    storage.terminate_cluster(id=cluster.id)


## Node Type ops:

def get_node_type(**args):
    return _node_type(storage.get_node_type(**args))


def get_node_types(**args):
    return [_node_type(t) for t in storage.get_node_types(**args)]


def get_node_type_required_params(**args):
    result = {}
    for process in storage.get_node_type(**args).processes:
        result[process.name] = []
        for prop in process.node_process_properties:
            if prop.required and not prop.default:
                result[process.name] += [prop.name]

    return result


def get_node_type_all_params(**args):
    result = {}
    for process in storage.get_node_type(**args).processes:
        result[process.name] = [prop.name
                                for prop in process.node_process_properties]
    return result


## Utils and DB object to Resource converters

def _clean_nones(obj):
    if not isinstance(obj, dict) and not isinstance(obj, list):
        return obj

    if isinstance(obj, dict):
        remove = []
        for key, value in obj.iteritems():
            if value is None:
                remove.append(key)
        for key in remove:
            obj.pop(key)
        for value in obj.values():
            _clean_nones(value)
    elif isinstance(obj, list):
        new_list = []
        for elem in obj:
            elem = _clean_nones(elem)
            if elem is not None:
                new_list.append(elem)
        return new_list

    return obj


class Resource(object):
    def __init__(self, _name, _info):
        self._name = _name
        self._info = _clean_nones(_info)

    def __getattr__(self, k):
        if k not in self.__dict__:
            return self._info.get(k)
        return self.__dict__[k]

    def __repr__(self):
        return '<%s %s>' % (self._name, self._info)

    def __eq__(self, other):
        return self._name == other._name and self._info == other._info

    @property
    def dict(self):
        return self._info

    @property
    def wrapped_dict(self):
        return {self._name: self._info}


def _node_template(nt):
    if not nt:
        raise ex.NodeTemplateNotFoundException(nt)

    d = {
        'id': nt.id,
        'name': nt.name,
        'node_type': {
            'name': nt.node_type.name,
            'processes': [p.name for p in nt.node_type.processes]},
        'flavor_id': nt.flavor_id
    }

    for conf in nt.node_template_configs:
        c_section = conf.node_process_property.node_process.name
        c_name = conf.node_process_property.name
        c_value = conf.value
        if c_section not in d:
            d[c_section] = dict()
        d[c_section][c_name] = c_value

    return Resource('node_template', d)


def _cluster(cluster):
    if not cluster:
        raise ex.ClusterNotFoundException(cluster)

    d = {
        'id': cluster.id,
        'name': cluster.name,
        'base_image_id': cluster.base_image_id,
        'status': cluster.status,
        'service_urls': {},
        'node_templates': {},
        'nodes': [{'vm_id': n.vm_id,
                   'node_template': {
                       'id': n.node_template.id,
                       'name': n.node_template.name
                   }}
                  for n in cluster.nodes]
    }
    for ntc in cluster.node_counts:
        d['node_templates'][ntc.node_template.name] = ntc.count

    for service in cluster.service_urls:
        d['service_urls'][service.name] = service.url

    return Resource('cluster', d)


def _node_type(nt):
    if not nt:
        raise ex.NodeTypeNotFoundException(nt)

    d = {
        'id': nt.id,
        'name': nt.name,
        'processes': [p.name for p in nt.processes]
    }

    return Resource('node_type', d)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mock import patch, Mock
from oslo.config import cfg
import unittest

from savanna.exceptions import NotFoundException, SavannaException
import savanna.openstack.common.exception as os_ex
from savanna.service.api import Resource
import savanna.service.validation as v


CONF = cfg.CONF
CONF.import_opt('allow_cluster_ops', 'savanna.config')


def _raise(ex):
    def function(*args, **kwargs):
        raise ex

    return function


def _cluster(base, **kwargs):
    base['cluster'].update(**kwargs)
    return base


def _template(base, **kwargs):
    base['node_template'].update(**kwargs)
    return base


class TestValidation(unittest.TestCase):
    def setUp(self):
        self._create_object_fun = None
        CONF.set_override('allow_cluster_ops', False)

    def tearDown(self):
        self._create_object_fun = None
        CONF.clear_override('allow_cluster_ops')

    @patch("savanna.utils.api.bad_request")
    @patch("savanna.utils.api.request_data")
    def test_malformed_request_body(self, request_data, bad_request):
        ex = os_ex.MalformedRequestBody()
        request_data.side_effect = _raise(ex)
        m_func = Mock()
        m_func.__name__ = "m_func"

        v.validate(m_func)(m_func)()

        self._assert_calls(bad_request,
                           (1, 'MALFORMED_REQUEST_BODY',
                            'Malformed message body: %(reason)s'))

    def _assert_exists_by_id(self, side_effect, assert_func=True):
        m_checker = Mock()
        m_checker.side_effect = side_effect
        m_func = Mock()
        m_func.__name__ = "m_func"

        v.exists_by_id(m_checker, "template_id")(m_func)(template_id="asd")

        m_checker.assert_called_once_with(id="asd")

        if assert_func:
            m_func.assert_called_once_with(template_id="asd")

    @patch("savanna.utils.api.internal_error")
    @patch("savanna.utils.api.not_found")
    def test_exists_by_id_passed(self, not_found, internal_error):
        self._assert_exists_by_id(None)

        self.assertEqual(not_found.call_count, 0)
        self.assertEqual(internal_error.call_count, 0)

    @patch("savanna.utils.api.internal_error")
    @patch("savanna.utils.api.not_found")
    def test_exists_by_id_failed(self, not_found, internal_error):
        self._assert_exists_by_id(_raise(NotFoundException("")), False)
        self.assertEqual(not_found.call_count, 1)
        self.assertEqual(internal_error.call_count, 0)

        self._assert_exists_by_id(_raise(SavannaException()), False)
        self.assertEqual(not_found.call_count, 1)
        self.assertEqual(internal_error.call_count, 1)

        self._assert_exists_by_id(_raise(AttributeError()), False)
        self.assertEqual(not_found.call_count, 1)
        self.assertEqual(internal_error.call_count, 2)

    def _assert_calls(self, mock, call_info):
        print "_assert_calls for %s, \n\t actual: %s , \n\t expected: %s" \
              % (mock, mock.call_args, call_info)
        if not call_info:
            self.assertEqual(mock.call_count, 0)
        else:
            self.assertEqual(mock.call_count, call_info[0])
            self.assertEqual(mock.call_args[0][0].code, call_info[1])
            self.assertEqual(mock.call_args[0][0].message, call_info[2])

    def _assert_create_object_validation(
            self, data, bad_req_i=None, not_found_i=None, int_err_i=None):

        request_data_p = patch("savanna.utils.api.request_data")
        bad_req_p = patch("savanna.utils.api.bad_request")
        not_found_p = patch("savanna.utils.api.not_found")
        int_err_p = patch("savanna.utils.api.internal_error")
        get_clusters_p = patch("savanna.service.api.get_clusters")
        get_templates_p = patch("savanna.service.api.get_node_templates")
        get_template_p = patch("savanna.service.api.get_node_template")
        get_types_p = patch("savanna.service.api.get_node_types")
        get_node_type_required_params_p = \
            patch("savanna.service.api.get_node_type_required_params")
        get_node_type_all_params_p = \
            patch("savanna.service.api.get_node_type_all_params")
        patchers = (request_data_p, bad_req_p, not_found_p, int_err_p,
                    get_clusters_p, get_templates_p, get_template_p,
                    get_types_p, get_node_type_required_params_p,
                    get_node_type_all_params_p)

        request_data = request_data_p.start()
        bad_req = bad_req_p.start()
        not_found = not_found_p.start()
        int_err = int_err_p.start()
        get_clusters = get_clusters_p.start()
        get_templates = get_templates_p.start()
        get_template = get_template_p.start()
        get_types = get_types_p.start()
        get_node_type_required_params = get_node_type_required_params_p.start()
        get_node_type_all_params = get_node_type_all_params_p.start()

        # stub clusters list
        get_clusters.return_value = getattr(self, "_clusters_data", [
            Resource("cluster", {
                "name": "some-cluster-1"
            })
        ])

        # stub node templates
        get_templates.return_value = getattr(self, "_templates_data", [
            Resource("node_template", {
                "name": "jt_nn.small",
                "node_type": {
                    "name": "JT+NN",
                    "processes": ["job_tracker", "name_node"]
                }
            }),
            Resource("node_template", {
                "name": "nn.small",
                "node_type": {
                    "name": "NN",
                    "processes": ["name_node"]
                }
            })
        ])

        def _get_template(name):
            for template in get_templates():
                if template.name == name:
                    return template
            return None

        get_template.side_effect = _get_template

        get_types.return_value = getattr(self, "_types_data", [
            Resource("node_type", {
                "name": "JT+NN",
                "processes": ["job_tracker", "name_node"]
            })
        ])

        def _get_r_params(name):
            if name == "JT+NN":
                return {"job_tracker": ["jt_param"]}
            return dict()

        get_node_type_required_params.side_effect = _get_r_params

        def _get_all_params(name):
            if name == "JT+NN":
                return {"job_tracker": ["jt_param"]}
            return dict()

        get_node_type_all_params.side_effect = _get_all_params

        # mock function that should be validated
        m_func = Mock()
        m_func.__name__ = "m_func"

        # request data to validate
        request_data.return_value = data

        v.validate(self._create_object_fun)(m_func)()

        self.assertEqual(request_data.call_count, 1)

        self._assert_calls(bad_req, bad_req_i)
        self._assert_calls(not_found, not_found_i)
        self._assert_calls(int_err, int_err_i)

        for patcher in patchers:
            patcher.stop()

    def test_cluster_create_v_required(self):
        self._create_object_fun = v.validate_cluster_create

        self._assert_create_object_validation(
            {},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'cluster' is a required property")
        )
        self._assert_create_object_validation(
            {"cluster": {}},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'name' is a required property")
        )
        self._assert_create_object_validation(
            {"cluster": {
                "name": "some-name"
            }},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'base_image_id' is a required property")
        )
        self._assert_create_object_validation(
            {"cluster": {
                "name": "some-name",
                "base_image_id": "some-image-id"
            }},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'node_templates' is a required property")
        )

    def test_cluster_create_v_name_base(self):
        self._create_object_fun = v.validate_cluster_create

        cluster = {
            "cluster": {
                "base_image_id": "some-image-id",
                "node_templates": {}
            }
        }
        self._assert_create_object_validation(
            _cluster(cluster, name=None),
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"None is not of type 'string'")
        )
        self._assert_create_object_validation(
            _cluster(cluster, name=""),
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'' is too short")
        )
        self._assert_create_object_validation(
            _cluster(cluster, name="a" * 51),
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'%s' is too long" % ('a' * 51))
        )

    def test_cluster_create_v_name_pattern(self):
        self._create_object_fun = v.validate_cluster_create

        cluster = {
            "cluster": {
                "base_image_id": "some-image-id",
                "node_templates": {}
            }
        }

        def _assert_cluster_name_pattern(self, name):
            cluster_schema = v.CLUSTER_CREATE_SCHEMA['properties']['cluster']
            name_p = cluster_schema['properties']['name']['pattern']
            self._assert_create_object_validation(
                _cluster(cluster, name=name),
                bad_req_i=(1, "VALIDATION_ERROR",
                           (u"'%s' does not match '%s'" % (name, name_p))
                           .replace('\\', "\\\\"))
            )

        _assert_cluster_name_pattern(self, "asd_123")
        _assert_cluster_name_pattern(self, "123")
        _assert_cluster_name_pattern(self, "asd?")

    def test_cluster_create_v_name_exists(self):
        self._create_object_fun = v.validate_cluster_create

        cluster = {
            "cluster": {
                "base_image_id": "some-image-id",
                "node_templates": {}
            }
        }

        self._assert_create_object_validation(
            _cluster(cluster, name="some-cluster-1"),
            bad_req_i=(1, "CLUSTER_NAME_ALREADY_EXISTS",
                       u"Cluster with name 'some-cluster-1' already exists")
        )

    def test_cluster_create_v_templates(self):
        self._create_object_fun = v.validate_cluster_create

        cluster = {
            "cluster": {
                "name": "some-cluster",
                "base_image_id": "some-image-id"
            }
        }
        self._assert_create_object_validation(
            _cluster(cluster, node_templates={}),
            bad_req_i=(1, "NOT_SINGLE_NAME_NODE",
                       u"Hadoop cluster should contain only 1 NameNode. "
                       u"Actual NN count is 0")
        )
        self._assert_create_object_validation(
            _cluster(cluster, node_templates={
                "nn.small": 1
            }),
            bad_req_i=(1, "NOT_SINGLE_JOB_TRACKER",
                       u"Hadoop cluster should contain only 1 JobTracker. "
                       u"Actual JT count is 0")
        )
        self._assert_create_object_validation(
            _cluster(cluster, node_templates={
                "incorrect_template": 10
            }),
            bad_req_i=(1, "NODE_TEMPLATE_NOT_FOUND",
                       u"NodeTemplate 'incorrect_template' not found")
        )
        self._assert_create_object_validation(
            _cluster(cluster, node_templates={
                "jt_nn.small": 1
            })
        )

    def test_node_template_create_v_required(self):
        self._create_object_fun = v.validate_node_template_create

        self._assert_create_object_validation(
            {},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'node_template' is a required property")
        )
        self._assert_create_object_validation(
            {"node_template": {}},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'name' is a required property")
        )
        self._assert_create_object_validation(
            {"node_template": {
                "name": "some-name"
            }},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'node_type' is a required property")
        )
        self._assert_create_object_validation(
            {"node_template": {
                "name": "some-name",
                "node_type": "some-node-type"
            }},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'flavor_id' is a required property")
        )
        self._assert_create_object_validation(
            {"node_template": {
                "name": "some-name",
                "node_type": "JT+NN",
                "flavor_id": "flavor-1"
            }},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'name_node' is a required property")
        )
        self._assert_create_object_validation(
            {"node_template": {
                "name": "some-name",
                "node_type": "JT+NN",
                "flavor_id": "flavor-1",
                "name_node": {}
            }},
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'job_tracker' is a required property")
        )
        self._assert_create_object_validation(
            {"node_template": {
                "name": "some-name",
                "node_type": "JT+NN",
                "flavor_id": "flavor-1",
                "name_node": {},
                "job_tracker": {}
            }},
            bad_req_i=(1, "REQUIRED_PARAM_MISSED",
                       u"Required parameter 'jt_param' of process "
                       u"'job_tracker' should be specified")
        )
        self._assert_create_object_validation(
            {"node_template": {
                "name": "some-name",
                "node_type": "JT+NN",
                "flavor_id": "flavor-1",
                "name_node": {},
                "job_tracker": {"jt_param": ""}
            }},
            bad_req_i=(1, "REQUIRED_PARAM_MISSED",
                       u"Required parameter 'jt_param' of process "
                       u"'job_tracker' should be specified")
        )
        self._assert_create_object_validation(
            {"node_template": {
                "name": "some-name",
                "node_type": "JT+NN",
                "flavor_id": "flavor-1",
                "name_node": {},
                "job_tracker": {"jt_param": "some value", "bad.parameter": "1"}
            }},
            bad_req_i=(1, "PARAM_IS_NOT_ALLOWED",
                       u"Parameter 'bad.parameter' "
                       u"of process 'job_tracker' is not allowed to change")
        )
        self._assert_create_object_validation(
            {"node_template": {
                "name": "some-name",
                "node_type": "JT+NN",
                "flavor_id": "flavor-1",
                "name_node": {},
                "job_tracker": {"jt_param": "some value"}
            }},
        )
        self._assert_create_object_validation(
            {"node_template": {
                "name": "some-name",
                "node_type": "JT+NN",
                "flavor_id": "flavor-1",
                "name_node": {},
                "job_tracker": {},
                "task_tracker": {}
            }},
            bad_req_i=(1, "NODE_PROCESS_DISCREPANCY",
                       u"Discrepancies in Node Processes. "
                       u"Required: ['name_node', 'job_tracker']")
        )

    def test_node_template_create_v_name_base(self):
        self._create_object_fun = v.validate_node_template_create

        template = {
            "node_template": {
                "node_type": "JT+NN",
                "flavor_id": "flavor-1",
                "name_node": {},
                "job_tracker": {}
            }
        }
        self._assert_create_object_validation(
            _template(template, name=None),
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"None is not of type 'string'")
        )
        self._assert_create_object_validation(
            _template(template, name=""),
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'' is too short")
        )
        self._assert_create_object_validation(
            _template(template, name="a" * 241),
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'%s' is too long" % ('a' * 241))
        )

    def test_node_template_create_v_name_pattern(self):
        self._create_object_fun = v.validate_node_template_create

        template = {
            "node_template": {
                "node_type": "JT+NN",
                "flavor_id": "flavor-1",
                "name_node": {},
                "job_tracker": {}
            }
        }

        def _assert_template_name_pattern(self, name):
            schema_props = v.TEMPLATE_CREATE_SCHEMA['properties']
            template_schema = schema_props['node_template']
            name_p = template_schema['properties']['name']['pattern']
            self._assert_create_object_validation(
                _template(template, name=name),
                bad_req_i=(1, "VALIDATION_ERROR",
                           (u"'%s' does not match '%s'" % (name, name_p))
                           .replace('\\', "\\\\"))
            )

        _assert_template_name_pattern(self, "asd;123")
        _assert_template_name_pattern(self, "123")
        _assert_template_name_pattern(self, "asd?")

    def test_node_template_create_v_name_exists(self):
        self._create_object_fun = v.validate_node_template_create

        template = {
            "node_template": {
                "node_type": "JT+NN",
                "flavor_id": "flavor-1",
                "name_node": {},
                "job_tracker": {}
            }
        }

        self._assert_create_object_validation(
            _template(template, name="jt_nn.small"),
            bad_req_i=(1, "NODE_TEMPLATE_ALREADY_EXISTS",
                       u"NodeTemplate with name 'jt_nn.small' already exists")
        )

    def test_node_template_create_v_types(self):
        self._create_object_fun = v.validate_node_template_create

        self._assert_create_object_validation(
            {
                "node_template": {
                    "name": "some-name",
                    "node_type": "JJ",
                    "flavor_id": "flavor-1",
                    "name_node": {},
                    "job_tracker": {}
                }
            },
            bad_req_i=(1, "NODE_TYPE_NOT_FOUND",
                       u"NodeType 'JJ' not found")
        )

# TODO(slukjanov): add tests for allow_cluster_ops = True



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo.config import cfg

from savanna.openstack.common import jsonutils
from savanna.openstack.common import log as logging


CONF = cfg.CONF


def notify(_context, message):
    """Notifies the recipient of the desired event given the model.
    Log notifications using openstack's default logging system"""

    priority = message.get('priority',
                           CONF.default_notification_level)
    priority = priority.lower()
    logger = logging.getLogger(
        'savanna.openstack.common.notification.%s' %
        message['event_type'])
    getattr(logger, priority)(jsonutils.dumps(message))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from novaclient.v1_1 import client as nova_client

import savanna.utils.openstack.base as base


def novaclient(headers):
    username = headers['X-User-Name']
    token = headers['X-Auth-Token']
    tenant = headers['X-Tenant-Id']
    compute_url = base.url_for(headers, 'compute')

    logging.debug('novaclient connection created using token '
                  '"%s", tenant "%s" and url "%s"',
                  token, tenant, compute_url)

    nova = nova_client.Client(username, token, tenant,
                              auth_url=compute_url)

    nova.client.auth_token = token
    nova.client.management_url = compute_url

    return nova


def get_flavors(headers):
    flavors = [flavor.name for flavor
               in novaclient(headers).flavors.list()]
    return flavors


def get_flavor(headers, **kwargs):
    return novaclient(headers).flavors.find(**kwargs)


def get_images(headers):
    images = [image.id for image
              in novaclient(headers).images.list()]
    return images


def get_limits(headers):
    limits = novaclient(headers).limits.get().absolute
    return dict((l.name, l.value) for l in limits)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
import json
import unittest

from savanna.openstack.common import log as logging
from savanna.tests.unit.base import SavannaTestCase

LOG = logging.getLogger(__name__)


class TestApiV02(SavannaTestCase):

    def test_list_node_templates(self):
        rv = self.app.get('/v0.2/some-tenant-id/node-templates.json')
        self.assertEquals(rv.status_code, 200)
        data = json.loads(rv.data)

        # clean all ids
        for idx in xrange(0, len(data.get(u'node_templates'))):
            del data.get(u'node_templates')[idx][u'id']

        self.assertEquals(data, _get_templates_stub_data())

    def test_create_node_template(self):
        rv = self.app.post('/v0.2/some-tenant-id/node-templates.json',
                           data=json.dumps(dict(
                               node_template=dict(
                                   name='test-template',
                                   node_type='JT+NN',
                                   flavor_id='test_flavor',
                                   job_tracker={
                                       'heap_size': '1234'
                                   },
                                   name_node={
                                       'heap_size': '2345'
                                   }
                               ))))
        self.assertEquals(rv.status_code, 202)
        data = json.loads(rv.data)

        data = data['node_template']

        # clean all ids
        del data[u'id']

        self.assertEquals(data, {
            u'job_tracker': {
                u'heap_size': u'1234'
            }, u'name': u'test-template',
            u'node_type': {
                u'processes': [
                    u'job_tracker', u'name_node'
                ],
                u'name': u'JT+NN'
            },
            u'flavor_id': u'test_flavor',
            u'name_node': {
                u'heap_size': u'2345'
            }
        })

    def test_list_clusters(self):
        rv = self.app.get('/v0.2/some-tenant-id/clusters.json')
        self.assertEquals(rv.status_code, 200)
        data = json.loads(rv.data)

        self.assertEquals(data, {
            u'clusters': []
        })

    def test_create_clusters(self):
        rv = self.app.post('/v0.2/some-tenant-id/clusters.json',
                           data=json.dumps(dict(
                               cluster=dict(
                                   name='test-cluster',
                                   base_image_id='base-image-id',
                                   node_templates={
                                       'jt_nn.medium': 1,
                                       'tt_dn.small': 5
                                   }
                               ))))
        self.assertEquals(rv.status_code, 202)
        data = json.loads(rv.data)

        data = data['cluster']

        cluster_id = data.pop(u'id')

        self.assertEquals(data, {
            u'status': u'Starting',
            u'service_urls': {},
            u'name': u'test-cluster',
            u'base_image_id': u'base-image-id',
            u'node_templates': {
                u'jt_nn.medium': 1,
                u'tt_dn.small': 5
            },
            u'nodes': []
        })

        eventlet.sleep(4)

        rv = self.app.get('/v0.2/some-tenant-id/clusters/%s.json' % cluster_id)
        self.assertEquals(rv.status_code, 200)
        data = json.loads(rv.data)

        data = data['cluster']

        self.assertEquals(data.pop(u'id'), cluster_id)

        # clean all ids
        for idx in xrange(0, len(data.get(u'nodes'))):
            del data.get(u'nodes')[idx][u'vm_id']
            del data.get(u'nodes')[idx][u'node_template'][u'id']

        nodes = data.pop(u'nodes')

        self.assertEquals(data, {
            u'status': u'Active',
            u'service_urls': {},
            u'name': u'test-cluster',
            u'base_image_id': u'base-image-id',
            u'node_templates': {
                u'jt_nn.medium': 1,
                u'tt_dn.small': 5
            }
        })

        self.assertEquals(_sorted_nodes(nodes), _sorted_nodes([
            {u'node_template': {u'name': u'tt_dn.small'}},
            {u'node_template': {u'name': u'tt_dn.small'}},
            {u'node_template': {u'name': u'tt_dn.small'}},
            {u'node_template': {u'name': u'tt_dn.small'}},
            {u'node_template': {u'name': u'tt_dn.small'}},
            {u'node_template': {u'name': u'jt_nn.medium'}}
        ]))

    def test_delete_node_template(self):
        rv = self.app.post('/v0.2/some-tenant-id/node-templates.json',
                           data=json.dumps(dict(
                               node_template=dict(
                                   name='test-template-2',
                                   node_type='JT+NN',
                                   flavor_id='test_flavor_2',
                                   job_tracker={
                                       'heap_size': '1234'
                                   },
                                   name_node={
                                       'heap_size': '2345'
                                   }
                               ))))
        self.assertEquals(rv.status_code, 202)
        data = json.loads(rv.data)

        data = data['node_template']

        node_template_id = data.pop(u'id')

        rv = self.app.get(
            '/v0.2/some-tenant-id/node-templates/%s.json' % node_template_id)
        self.assertEquals(rv.status_code, 200)
        data = json.loads(rv.data)

        data = data['node_template']

        # clean all ids
        del data[u'id']

        self.assertEquals(data, {
            u'job_tracker': {
                u'heap_size': u'1234'
            }, u'name': u'test-template-2',
            u'node_type': {
                u'processes': [
                    u'job_tracker', u'name_node'
                ],
                u'name': u'JT+NN'
            },
            u'flavor_id': u'test_flavor_2',
            u'name_node': {
                u'heap_size': u'2345'
            }
        })

        rv = self.app.delete(
            '/v0.2/some-tenant-id/node-templates/%s.json' % node_template_id)
        self.assertEquals(rv.status_code, 204)

        rv = self.app.get(
            '/v0.2/some-tenant-id/node-templates/%s.json' % node_template_id)
        self.assertEquals(rv.status_code, 404)

    def test_delete_cluster(self):
        rv = self.app.post('/v0.2/some-tenant-id/clusters.json',
                           data=json.dumps(dict(
                               cluster=dict(
                                   name='test-cluster-2',
                                   base_image_id='base-image-id_2',
                                   node_templates={
                                       'jt_nn.medium': 1,
                                       'tt_dn.small': 5
                                   }
                               ))))
        self.assertEquals(rv.status_code, 202)
        data = json.loads(rv.data)

        data = data['cluster']

        cluster_id = data.pop(u'id')

        rv = self.app.get('/v0.2/some-tenant-id/clusters/%s.json' % cluster_id)
        self.assertEquals(rv.status_code, 200)
        data = json.loads(rv.data)

        data = data['cluster']

        # delete all ids
        del data[u'id']

        self.assertEquals(data, {
            u'status': u'Starting',
            u'service_urls': {},
            u'name': u'test-cluster-2',
            u'base_image_id': u'base-image-id_2',
            u'node_templates': {
                u'jt_nn.medium': 1,
                u'tt_dn.small': 5
            },
            u'nodes': []
        })

        rv = self.app.delete(
            '/v0.2/some-tenant-id/clusters/%s.json' % cluster_id)
        self.assertEquals(rv.status_code, 204)

        eventlet.sleep(1)

        rv = self.app.get('/v0.2/some-tenant-id/clusters/%s.json' % cluster_id)
        self.assertEquals(rv.status_code, 404)


def _sorted_nodes(nodes):
    return sorted(nodes, key=lambda elem: elem[u'node_template'][u'name'])


def _get_templates_stub_data():
    return {
        u'node_templates': [
            {
                u'job_tracker': {
                    u'heap_size': u'896'
                },
                u'name': u'jt_nn.small',
                u'node_type': {
                    u'processes': [
                        u'job_tracker', u'name_node'
                    ],
                    u'name': u'JT+NN'
                },
                u'flavor_id': u'm1.small',
                u'name_node': {
                    u'heap_size': u'896'
                }
            },
            {
                u'job_tracker': {
                    u'heap_size': u'1792'
                },
                u'name': u'jt_nn.medium',
                u'node_type': {
                    u'processes': [
                        u'job_tracker', u'name_node'
                    ], u'name': u'JT+NN'
                },
                u'flavor_id': u'm1.medium',
                u'name_node': {
                    u'heap_size': u'1792'
                }
            },
            {
                u'name': u'tt_dn.small',
                u'task_tracker': {
                    u'heap_size': u'896'
                },
                u'data_node': {
                    u'heap_size': u'896'
                },
                u'node_type': {
                    u'processes': [
                        u'task_tracker', u'data_node'
                    ],
                    u'name': u'TT+DN'
                },
                u'flavor_id': u'm1.small'
            },
            {
                u'name': u'tt_dn.medium',
                u'task_tracker': {
                    u'heap_size': u'1792',
                },
                u'data_node': {
                    u'heap_size': u'1792'
                },
                u'node_type': {
                    u'processes': [
                        u'task_tracker', u'data_node'
                    ],
                    u'name': u'TT+DN'
                },
                u'flavor_id': u'm1.medium'
            }
        ]
    }


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
from jinja2 import Environment
from jinja2 import PackageLoader
from oslo.config import cfg
import paramiko
from pkg_resources import resource_filename
import xml.dom.minidom as xml

from savanna.openstack.common import log as logging
from savanna.storage.db import DB
from savanna.storage.models import Node, ServiceUrl
from savanna.storage.storage import update_cluster_status
from savanna.utils.openstack.nova import novaclient
from savanna.utils.patches import patch_minidom_writexml


LOG = logging.getLogger(__name__)

CONF = cfg.CONF

cluster_node_opts = [
    cfg.StrOpt('username',
               default='root',
               help='An existing user on Hadoop image'),
    cfg.StrOpt('password',
               default='swordfish',
               help='User\'s password'),
    cfg.BoolOpt('use_floating_ips',
                default=True,
                help='When set to false, Savanna uses only internal IP of VMs.'
                     ' When set to true, Savanna expects OpenStack to auto-'
                     'assign floating IPs to cluster nodes. Internal IPs will '
                     'be used for inter-cluster communication, while floating '
                     'ones will be used by Savanna to configure nodes. Also '
                     'floating IPs will be exposed in service URLs')
]

CONF.register_opts(cluster_node_opts, 'cluster_node')


def _find_by_id(lst, id):
    for entity in lst:
        if entity.id == id:
            return entity

    return None


def _find_by_name(lst, name):
    for entity in lst:
        if entity.name == name:
            return entity

    return None


def _check_finding(entity, attr, value):
    if entity is None:
        raise RuntimeError("Unable to find entity with %s "
                           "\'%s\'" % (attr, value))


def _ensure_zero(ret):
    if ret != 0:
        raise RuntimeError('Command returned non-zero status code - %i' % ret)


def _setup_ssh_connection(host, ssh):
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(
        host,
        username=CONF.cluster_node.username,
        password=CONF.cluster_node.password
    )


def _open_channel_and_execute(ssh, cmd):
    chan = ssh.get_transport().open_session()
    chan.exec_command(cmd)
    return chan.recv_exit_status()


def _execute_command_on_node(host, cmd):
    ssh = paramiko.SSHClient()
    try:
        _setup_ssh_connection(host, ssh)
        return _open_channel_and_execute(ssh, cmd)
    finally:
        ssh.close()


def launch_cluster(headers, cluster):
    nova = novaclient(headers)

    clmap = dict()
    clmap['id'] = cluster.id
    clmap['name'] = cluster.name
    clmap['image'] = _find_by_id(nova.images.list(),
                                 cluster.base_image_id)
    _check_finding(clmap['image'], 'id', cluster.base_image_id)

    clmap['nodes'] = []
    num = 1

    for nc in cluster.node_counts:
        configs = dict()
        for cf in nc.node_template.node_template_configs:
            proc_name = cf.node_process_property.node_process.name
            if proc_name not in configs:
                configs[proc_name] = dict()

            name = cf.node_process_property.name
            configs[proc_name][name] = cf.value

        ntype = nc.node_template.node_type.name
        templ_id = nc.node_template.id
        flv_id = nc.node_template.flavor_id
        flv = _find_by_name(nova.flavors.list(), flv_id)
        _check_finding(flv, 'id', flv_id)

        for _ in xrange(0, nc.count):
            node = dict()
            node['id'] = None
            if ntype == 'JT+NN':
                node['name'] = '%s-master' % cluster.name
            else:
                node['name'] = '%s-%i' % (cluster.name, num)
                num += 1
            node['type'] = ntype
            node['templ_id'] = templ_id
            node['flavor'] = flv
            node['configs'] = configs
            node['is_up'] = False
            clmap['nodes'].append(node)

    try:
        for node in clmap['nodes']:
            LOG.debug("Starting node for cluster '%s', node: %s, image: %s",
                      cluster.name, node, clmap['image'])
            _launch_node(nova, node, clmap['image'])
    except Exception, e:
        _rollback_cluster_creation(cluster, clmap, nova, e)
        return False

    all_set = False

    LOG.debug("All nodes for cluster '%s' have been started, "
              "waiting for them to come up", cluster.name)

    while not all_set:
        all_set = True

        for node in clmap['nodes']:
            _check_if_up(nova, node)

            if not node['is_up']:
                all_set = False

        eventlet.sleep(1)

    LOG.debug("All nodes of cluster '%s' are up: %s",
              cluster.name, all_set)

    _pre_cluster_setup(clmap)
    for node in clmap['nodes']:
        _setup_node(node, clmap)
        _register_node(node, cluster)

    LOG.debug("All nodes of cluster '%s' are configured and registered, "
              "starting the cluster...", cluster.name)

    _start_cluster(cluster, clmap)

    return True


def _launch_node(nova, node, image):
    srv = nova.servers.create(node['name'], image, node['flavor'])
    node['id'] = srv.id


def _rollback_cluster_creation(cluster, clmap, nova, error):
    update_cluster_status("Error", id=cluster.id)

    LOG.warn("Can't launch all vms for cluster '%s': %s", cluster.id, error)
    for node in clmap['nodes']:
        if node['id']:
            _stop_node_silently(nova, cluster, node['id'])

    LOG.info("All vms of cluster '%s' has been stopped", cluster.id)


def _stop_node_silently(nova, cluster, vm_id):
    LOG.debug("Stopping vm '%s' of cluster '%s'", vm_id, cluster.id)
    try:
        nova.servers.delete(vm_id)
    except Exception, e:
        LOG.error("Can't silently remove node '%s': %s", vm_id, e)


def _check_if_up(nova, node):
    if node['is_up']:
        # all set
        return

    if 'ip' not in node:
        srv = _find_by_id(nova.servers.list(), node['id'])
        nets = srv.networks

        if len(nets) == 0:
            # VM's networking is not configured yet
            return

        ips = nets.values()[0]

        if CONF.cluster_node.use_floating_ips:
            if len(ips) < 2:
                # floating IP is not assigned yet
                return

            # we assume that floating IP comes last in the list
            node['ip'] = ips[-1]
            node['internal_ip'] = ips[0]
        else:
            if len(ips) < 1:
                # private IP is not assigned yet
                return
            node['ip'] = ips[0]
            node['internal_ip'] = ips[0]

    try:
        ret = _execute_command_on_node(node['ip'], 'ls -l /')
        _ensure_zero(ret)
    except Exception:
        # ssh is not up yet
        return

    node['is_up'] = True


env = Environment(loader=PackageLoader('savanna', 'resources'))


def _render_template(template_name, **kwargs):
    templ = env.get_template('%s.template' % template_name)
    return templ.render(**kwargs)


ENV_CONFS = {
    'job_tracker': {
        'heap_size': 'HADOOP_JOBTRACKER_OPTS=\\"-Xmx%sm\\"'
    },
    'name_node': {
        'heap_size': 'HADOOP_NAMENODE_OPTS=\\"-Xmx%sm\\"'
    },
    'task_tracker': {
        'heap_size': 'HADOOP_TASKTRACKER_OPTS=\\"-Xmx%sm\\"'
    },
    'data_node': {
        'heap_size': 'HADOOP_DATANODE_OPTS=\\"-Xmx%sm\\"'
    }
}


def _load_xml_default_configs(file_name):
    doc = xml.parse(resource_filename("savanna", 'resources/%s' % file_name))
    properties = doc.getElementsByTagName("name")
    return [prop.childNodes[0].data for prop in properties]

CORE_CONF = _load_xml_default_configs('core-default.xml')
MAPRED_CONF = _load_xml_default_configs('mapred-default.xml')
HDFS_CONF = _load_xml_default_configs('hdfs-default.xml')


def _generate_xml_configs(node, clmap):
    # inserting common configs depends on provisioned VMs and HDFS placement
    cfg = {
        'fs.default.name': 'hdfs://%s:8020' % clmap['master_hostname'],
        'mapred.job.tracker': '%s:8021' % clmap['master_hostname'],
        'dfs.name.dir': '/mnt/lib/hadoop/hdfs/namenode',
        'dfs.data.dir': '/mnt/lib/hadoop/hdfs/datanode',
        'mapred.system.dir': '/mnt/mapred/mapredsystem',
        'mapred.local.dir': '/mnt/lib/hadoop/mapred'
    }

    # inserting user-defined configs from NodeTemplates
    for key, value in _extract_xml_confs(node['configs']):
        cfg[key] = value

    # invoking applied configs to appropriate xml files
    xml_configs = {
        'core-site': _create_xml(cfg, CORE_CONF),
        'mapred-site': _create_xml(cfg, MAPRED_CONF),
        'hdfs-site': _create_xml(cfg, HDFS_CONF)
    }

    return xml_configs


# Patches minidom's writexml to avoid excess whitespaces in generated xml
# configuration files that brakes Hadoop.
patch_minidom_writexml()


def _create_xml(configs, global_conf):
    doc = xml.Document()

    pi = doc.createProcessingInstruction('xml-stylesheet',
                                         'type="text/xsl" '
                                         'href="configuration.xsl"')
    doc.insertBefore(pi, doc.firstChild)

    # Create the <configuration> base element
    configuration = doc.createElement("configuration")
    doc.appendChild(configuration)

    for prop_name, prop_value in configs.items():
        if prop_name in global_conf:
            # Create the <property> element
            property = doc.createElement("property")
            configuration.appendChild(property)

            # Create a <name> element in <property>
            name = doc.createElement("name")
            property.appendChild(name)

            # Give the <name> element some hadoop config name
            name_text = doc.createTextNode(prop_name)
            name.appendChild(name_text)

            # Create a <value> element in <property>
            value = doc.createElement("value")
            property.appendChild(value)

            # Give the <value> element some hadoop config value
            value_text = doc.createTextNode(prop_value)
            value.appendChild(value_text)

    # Return newly created XML
    return doc.toprettyxml(indent="  ")


def _keys_exist(map, key1, key2):
    return key1 in map and key2 in map[key1]


def _extract_environment_confs(node_configs):
    """Returns list of Hadoop parameters which should be passed via environment
    """

    lst = []

    for process, proc_confs in ENV_CONFS.items():
        for param_name, param_format_str in proc_confs.items():
            if (_keys_exist(node_configs, process, param_name) and
                    not node_configs[process][param_name] is None):
                lst.append(param_format_str %
                           node_configs[process][param_name])

    return lst


def _extract_xml_confs(node_configs):
    """Returns list of Hadoop parameters which should be passed into general
    configs like core-site.xml
    """

    # For now we assume that all parameters outside of ENV_CONFS
    # are passed to xml files

    lst = []

    for process, proc_confs in node_configs.items():
        for param_name, param_value in proc_confs.items():
            if (not _keys_exist(ENV_CONFS, process, param_name) and
                    not param_value is None):
                lst.append((param_name, param_value))

    return lst


def _analyze_templates(clmap):
    clmap['master_ip'] = None
    clmap['slaves'] = []
    for node in clmap['nodes']:
        if node['type'] == 'JT+NN':
            clmap['master_ip'] = node['ip']
            clmap['master_hostname'] = node['name']
            node['is_master'] = True
        elif node['type'] == 'TT+DN':
            clmap['slaves'].append(node['name'])
            node['is_master'] = False

    if clmap['master_ip'] is None:
        raise RuntimeError("No master node is defined in the cluster")


def _generate_hosts(clmap):
    hosts = "127.0.0.1 localhost\n"
    for node in clmap['nodes']:
        hosts += "%s %s\n" % (node['internal_ip'], node['name'])

    clmap['hosts'] = hosts


def _pre_cluster_setup(clmap):
    _analyze_templates(clmap)
    _generate_hosts(clmap)

    for node in clmap['nodes']:
        if node['is_master']:
            script_file = 'setup-master.sh'
        else:
            script_file = 'setup-general.sh'

        templ_args = {
            'slaves': clmap['slaves'],
            'master_hostname': clmap['master_hostname'],
            'env_configs': _extract_environment_confs(node['configs'])
        }

        node['setup_script'] = _render_template(script_file, **templ_args)
        node['xml'] = _generate_xml_configs(node, clmap)


def _setup_node(node, clmap):
    ssh = paramiko.SSHClient()
    try:
        _setup_ssh_connection(node['ip'], ssh)
        sftp = ssh.open_sftp()

        fl = sftp.file('/etc/hosts', 'w')
        fl.write(clmap['hosts'])
        fl.close()

        fl = sftp.file('/etc/hadoop/core-site.xml', 'w')
        fl.write(node['xml']['core-site'])
        fl.close()

        fl = sftp.file('/etc/hadoop/hdfs-site.xml', 'w')
        fl.write(node['xml']['hdfs-site'])
        fl.close()

        fl = sftp.file('/etc/hadoop/mapred-site.xml', 'w')
        fl.write(node['xml']['mapred-site'])
        fl.close()

        fl = sftp.file('/tmp/savanna-hadoop-init.sh', 'w')
        fl.write(node['setup_script'])
        fl.close()

        sftp.chmod('/tmp/savanna-hadoop-init.sh', 0500)

        ret = _open_channel_and_execute(ssh,
                                        '/tmp/savanna-hadoop-init.sh '
                                        '>> /tmp/savanna-hadoop-init.log 2>&1')
        _ensure_zero(ret)
    finally:
        ssh.close()


def _register_node(node, cluster):
    node_obj = Node(node['id'], cluster.id, node['templ_id'])
    DB.session.add(node_obj)

    if node['is_master']:
        srv_url_jt = ServiceUrl(cluster.id, 'jobtracker', 'http://%s:50030'
                                                          % node['ip'])
        srv_url_nn = ServiceUrl(cluster.id, 'namenode', 'http://%s:50070'
                                                        % node['ip'])

        DB.session.add(srv_url_jt)
        DB.session.add(srv_url_nn)

    DB.session.commit()


def _start_cluster(cluster, clmap):
    ret = _execute_command_on_node(
        clmap['master_ip'],
        'su -c start-all.sh hadoop >> /tmp/savanna-hadoop-start-all.log')
    _ensure_zero(ret)

    LOG.info("Cluster '%s' successfully started!", cluster.name)


def stop_cluster(headers, cluster):
    nova = novaclient(headers)

    for node in cluster.nodes:
        try:
            nova.servers.delete(node.vm_id)
            LOG.debug("vm '%s' has been stopped", node.vm_id)
        except Exception, e:
            LOG.info("Can't stop vm '%s': %s", node.vm_id, e)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mock import patch
import unittest

import savanna.service.api as api


class TestServiceLayer(unittest.TestCase):
    ## Node Template ops:

    @patch('savanna.storage.storage.get_node_template')
    def test_get_node_template(self, m):
        m.return_value = api.Resource("node_template", {
            "id": "template-id",
            "name": "jt_nn.small",
            "node_type": api.Resource("node_type", {
                "name": "JT+NN",
                "processes": [
                    api.Resource("process", {"name": "job_tracker"}),
                    api.Resource("process", {"name": "name_node"})
                ]
            }),
            "flavor_id": "flavor-id",
            "node_template_configs": [
                api.Resource("conf", {
                    "node_process_property": api.Resource("prop", {
                        "name": "heap_size",
                        "node_process": api.Resource("process", {
                            "name": "job_tracker"
                        })
                    }),
                    "value": "1234"
                }),
                api.Resource("conf", {
                    "node_process_property": api.Resource("prop", {
                        "name": "heap_size",
                        "node_process": api.Resource("process", {
                            "name": "name_node"
                        })
                    }),
                    "value": "5678"
                })
            ]
        })

        nt = api.get_node_template(id='template-id')
        self.assertEqual(nt, api.Resource("node_template", {
            'id': 'template-id',
            'name': 'jt_nn.small',
            'node_type': {
                'processes': ['job_tracker', 'name_node'],
                'name': 'JT+NN'
            },
            'flavor_id': 'flavor-id',
            'job_tracker': {'heap_size': '1234'},
            'name_node': {'heap_size': '5678'}
        }))
        m.assert_called_once_with(id='template-id')

    @patch('savanna.storage.storage.get_node_templates')
    def test_get_node_templates(self, m):
        # '_node_template' tested in 'test_get_node_template'
        api.get_node_templates(node_type='JT+NN')
        m.assert_called_once_with(node_type='JT+NN')

    @patch('savanna.service.api.get_node_template')
    @patch('savanna.storage.storage.create_node_template')
    @patch('savanna.storage.storage.get_node_type')
    def test_create_node_template(self, get_n_type, create_tmpl, get_tmpl):
        get_n_type.return_value = api.Resource(
            "node_type", {"id": "node-type-1"})
        create_tmpl.return_value = api.Resource(
            "node-template", {"id": "tmpl-1"})

        api.create_node_template(
            {
                "node_template": {
                    "name": "nt-1",
                    "node_type": "JT+NN",
                    "flavor_id": "flavor-1"
                }
            }, {"X-Tenant-Id": "tenant-01"})

        get_n_type.assert_called_once_with(name="JT+NN")
        create_tmpl.assert_called_once_with("nt-1", "node-type-1",
                                            "flavor-1", {})
        get_tmpl.assert_called_once_with(id="tmpl-1")

    @patch('savanna.storage.storage.terminate_node_template')
    def test_terminate_node_template(self, m):
        api.terminate_node_template(node_type='JT+NN')
        m.assert_called_once_with(node_type='JT+NN')

    ## Cluster ops:

    @patch('savanna.storage.storage.get_cluster')
    def test_get_cluster(self, m):
        m.return_value = api.Resource("cluster", {
            "id": "cluster-id",
            "name": "cluster-name",
            "base_image_id": "image-id",
            "status": "Active",
            "nodes": [
                api.Resource("node", {
                    "vm_id": "vm-1",
                    "node_template": api.Resource("node_template", {
                        "id": "jt_nn.small-id",
                        "name": "jt_nn.small"
                    })
                }),
                api.Resource("node", {
                    "vm_id": "vm-2",
                    "node_template": api.Resource("node_template", {
                        "id": "tt_dn.small-id",
                        "name": "tt_dn.small"
                    })
                }),
                api.Resource("node", {
                    "vm_id": "vm-3",
                    "node_template": api.Resource("node_template", {
                        "id": "tt_dn.small-id",
                        "name": "tt_dn.small"
                    })
                })
            ],
            "node_counts": [
                api.Resource("node_count", {
                    "node_template": api.Resource("node_template", {
                        "name": "jt_nn.small"
                    }),
                    "count": "1"
                }),
                api.Resource("node_count", {
                    "node_template": api.Resource("node_template", {
                        "name": "tt_dn.small"
                    }),
                    "count": "2"
                })
            ],
            "service_urls": [
                api.Resource("service_url", {
                    "name": "job_tracker",
                    "url": "some-url"
                }),
                api.Resource("service_url", {
                    "name": "name_node",
                    "url": "some-url-2"
                })
            ]
        })

        cluster = api.get_cluster(id="cluster-id")
        self.assertEqual(cluster, api.Resource("cluster", {
            'id': 'cluster-id',
            'name': 'cluster-name',
            'base_image_id': "image-id",
            'status': 'Active',
            'node_templates': {'jt_nn.small': '1', 'tt_dn.small': '2'},
            'nodes': [
                {
                    'node_template': {
                        'id': 'jt_nn.small-id', 'name': 'jt_nn.small'
                    }, 'vm_id': 'vm-1'
                },
                {
                    'node_template': {
                        'id': 'tt_dn.small-id', 'name': 'tt_dn.small'
                    }, 'vm_id': 'vm-2'
                },
                {
                    'node_template': {
                        'id': 'tt_dn.small-id', 'name': 'tt_dn.small'
                    }, 'vm_id': 'vm-3'
                }
            ],
            'service_urls': {
                'name_node': 'some-url-2',
                'job_tracker': 'some-url'
            }
        }))
        m.assert_called_once_with(id="cluster-id")

    @patch('savanna.storage.storage.get_clusters')
    def test_get_clusters(self, m):
        # '_clusters' tested in 'test_get_clusters'
        api.get_clusters(id="cluster-id")
        m.assert_called_once_with(id="cluster-id")

    @patch('eventlet.spawn')
    @patch('savanna.service.api.get_cluster')
    @patch('savanna.storage.storage.create_cluster')
    def test_create_cluster(self, create_c, get_c, spawn):
        create_c.return_value = api.Resource("cluster", {
            "id": "cluster-1"
        })

        api.create_cluster(
            {
                "cluster": {
                    "name": "cluster-1",
                    "base_image_id": "image-1",
                    "node_templates": {
                        "jt_nn.small": "1",
                        "tt_dn.small": "10"
                    }
                }
            }, {"X-Tenant-Id": "tenant-01"})

        create_c.assert_called_once_with("cluster-1", "image-1", "tenant-01", {
            "jt_nn.small": "1",
            "tt_dn.small": "10"
        })
        get_c.assert_called_once_with(id="cluster-1")
        spawn.assert_called_once_with(api._cluster_creation_job,
                                      {"X-Tenant-Id": "tenant-01"},
                                      "cluster-1")

    @patch('eventlet.spawn')
    @patch('savanna.storage.storage.update_cluster_status')
    def test_terminate_cluster(self, update_status, spawn):
        update_status.return_value = api.Resource("cluster", {
            "id": "cluster-id"
        })

        api.terminate_cluster({"X-Tenant-Id": "tenant-01"}, id="cluster-id")

        update_status.assert_called_once_with('Stopping', id="cluster-id")
        spawn.assert_called_once_with(api._cluster_termination_job,
                                      {"X-Tenant-Id": "tenant-01"},
                                      "cluster-id")



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mimetypes
import traceback

from flask import abort, request, Blueprint, Response
from werkzeug.datastructures import MIMEAccept

from savanna.openstack.common.wsgi import JSONDictSerializer, \
    XMLDictSerializer, JSONDeserializer
from savanna.openstack.common import log as logging

LOG = logging.getLogger(__name__)


class Rest(Blueprint):
    def get(self, rule, status_code=200):
        return self._mroute('GET', rule, status_code)

    def post(self, rule, status_code=202):
        return self._mroute('POST', rule, status_code)

    def put(self, rule, status_code=202):
        return self._mroute('PUT', rule, status_code)

    def delete(self, rule, status_code=204):
        return self._mroute('DELETE', rule, status_code)

    def _mroute(self, methods, rule, status_code=None):
        if type(methods) is str:
            methods = [methods]
        return self.route(rule, methods=methods, status_code=status_code)

    def route(self, rule, **options):
        status = options.pop('status_code', None)

        def decorator(func):
            endpoint = options.pop('endpoint', func.__name__)

            def handler(**kwargs):
                # extract response content type
                resp_type = request.accept_mimetypes
                type_suffix = kwargs.pop('resp_type', None)
                if type_suffix:
                    suffix_mime = mimetypes.guess_type("res." + type_suffix)[0]
                    if suffix_mime:
                        resp_type = MIMEAccept([(suffix_mime, 1)])
                request.resp_type = resp_type

                # extract fields (column selection)
                fields = list(set(request.args.getlist('fields')))
                fields.sort()
                request.fields_selector = fields

                if status:
                    request.status_code = status

                kwargs.pop("tenant_id")

                return func(**kwargs)

            f_rule = "/<tenant_id>" + rule
            self.add_url_rule(f_rule, endpoint, handler, **options)
            ext_rule = f_rule + '.<resp_type>'
            self.add_url_rule(ext_rule, endpoint, handler, **options)

            return func

        return decorator


RT_JSON = MIMEAccept([("application/json", 1)])
RT_XML = MIMEAccept([("application/xml", 1)])


def render(res=None, resp_type=None, status=None, **kwargs):
    if not res:
        res = {}
    if type(res) is dict:
        res.update(kwargs)
    elif kwargs:
        # can't merge kwargs into the non-dict res
        abort_and_log(500, "Non-dict and non-empty kwargs passed to render")

    status_code = getattr(request, 'status_code', None)
    if status:
        status_code = status
    if not status_code:
        status_code = 200

    if not resp_type:
        resp_type = getattr(request, 'resp_type', RT_JSON)

    if not resp_type:
        resp_type = RT_JSON

    serializer = None
    if "application/json" in resp_type:
        resp_type = RT_JSON
        serializer = JSONDictSerializer()
    elif "application/xml" in resp_type:
        resp_type = RT_XML
        serializer = XMLDictSerializer()
    else:
        abort_and_log(400, "Content type '%s' isn't supported" % resp_type)

    body = serializer.serialize(res)
    resp_type = str(resp_type)

    return Response(response=body, status=status_code, mimetype=resp_type)


def request_data():
    if hasattr(request, 'parsed_data'):
        return request.parsed_data

    if not request.content_length > 0:
        LOG.debug("Empty body provided in request")
        return dict()

    deserializer = None
    content_type = request.mimetype
    if not content_type or content_type in RT_JSON:
        deserializer = JSONDeserializer()
    elif content_type in RT_XML:
        abort_and_log(400, "XML requests are not supported yet")
        # deserializer = XMLDeserializer()
    else:
        abort_and_log(400, "Content type '%s' isn't supported" % content_type)

    # parsed request data to avoid unwanted re-parsings
    request.parsed_data = deserializer.deserialize(request.data)['body']

    return request.parsed_data


def abort_and_log(status_code, descr, exc=None):
    LOG.error("Request aborted with status code %s and message '%s'",
              status_code, descr)

    if exc is not None:
        LOG.error(traceback.format_exc())

    abort(status_code, description=descr)


def render_error_message(error_code, error_message, error_name):
    message = {
        "error_code": error_code,
        "error_message": error_message,
        "error_name": error_name
    }

    resp = render(message)
    resp.status_code = error_code

    return resp


def internal_error(status_code, descr, exc=None):
    LOG.error("Request aborted with status code %s and message '%s'",
              status_code, descr)

    if exc is not None:
        LOG.error(traceback.format_exc())

    error_code = "INTERNAL_SERVER_ERROR"
    if status_code == 501:
        error_code = "NOT_IMPLEMENTED_ERROR"

    return render_error_message(status_code, descr, error_code)


def bad_request(error):
    error_code = 400

    LOG.debug("Validation Error occurred: "
              "error_code=%s, error_message=%s, error_name=%s",
              error_code, error.message, error.code)

    return render_error_message(error_code, error.message, error.code)


def not_found(error):
    error_code = 404

    LOG.debug("Not Found exception occurred: "
              "error_code=%s, error_message=%s, error_name=%s",
              error_code, error.message, error.code)

    return render_error_message(error_code, error.message, error.code)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Greenthread local storage of variables using weak references"""

import weakref

from eventlet import corolocal


class WeakLocal(corolocal.local):
    def __getattribute__(self, attr):
        rval = corolocal.local.__getattribute__(self, attr)
        if rval:
            # NOTE(mikal): this bit is confusing. What is stored is a weak
            # reference, not the value itself. We therefore need to lookup
            # the weak reference and return the inner value here.
            rval = rval()
        return rval

    def __setattr__(self, attr, value):
        value = weakref.ref(value)
        return corolocal.local.__setattr__(self, attr, value)


# NOTE(mikal): the name "store" should be deprecated in the future
store = WeakLocal()

# A "weak" store uses weak references and allows an object to fall out of scope
# when it falls out of scope in the code that uses the thread local storage. A
# "strong" store will hold a reference to the object so that it never falls out
# of scope.
weak_store = WeakLocal()
strong_store = corolocal.local



import setuptools

from savanna.openstack.common import setup as common_setup

requires = common_setup.parse_requirements()
depend_links = common_setup.parse_dependency_links()
project = 'savanna'

setuptools.setup(
    name=project,
    version=common_setup.get_version(project, '0.1.2'),
    description='Savanna project',
    author='Mirantis Inc.',
    author_email='savanna-team@mirantis.com',
    url='http://savanna.mirantis.com',
    classifiers=[
        'Environment :: OpenStack',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    license='Apache Software License',
    cmdclass=common_setup.get_cmdclass(),
    packages=setuptools.find_packages(exclude=['bin']),
    package_data={'savanna': [
        'resources/*.template',
        'resources/*.xml'
    ]},
    install_requires=requires,
    dependency_links=depend_links,
    setup_requires=['setuptools-git>=0.4'],
    include_package_data=True,
    test_suite='nose.collector',
    scripts=[
        'bin/savanna-api',
        'bin/savanna-manage',
    ],
    py_modules=[],
    data_files=[
        ('share/savanna',
         [
             'etc/savanna/savanna.conf.sample',
             'etc/savanna/savanna.conf.sample-full',
         ]),
    ],
)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from flask import request

from savanna.openstack.common import log as logging
from savanna.service import api
import savanna.service.validation as v
import savanna.utils.api as api_u

LOG = logging.getLogger(__name__)

rest = api_u.Rest('v02', __name__)


@rest.get('/node-templates')
def templates_list():
    try:
        return api_u.render(
            node_templates=[nt.dict for nt in api.get_node_templates()])
    except Exception, e:
        return api_u.internal_error(500,
                                    "Exception while listing NodeTemplates", e)


@rest.post('/node-templates')
@v.validate(v.validate_node_template_create)
def templates_create():
    data = api_u.request_data()
    headers = request.headers

    return api_u.render(api.create_node_template(data, headers).wrapped_dict)


@rest.get('/node-templates/<template_id>')
@v.exists_by_id(api.get_node_template, 'template_id')
def templates_get(template_id):
    nt = api.get_node_template(id=template_id)
    return api_u.render(nt.wrapped_dict)


@rest.put('/node-templates/<template_id>')
def templates_update(template_id):
    return api_u.internal_error(501, NotImplementedError(
        "Template update op isn't implemented (id '%s')"
        % template_id))


@rest.delete('/node-templates/<template_id>')
@v.exists_by_id(api.get_node_template, 'template_id')
@v.validate(v.validate_node_template_terminate)
def templates_delete(template_id):
    api.terminate_node_template(id=template_id)
    return api_u.render()


@rest.get('/clusters')
def clusters_list():
    tenant_id = request.headers['X-Tenant-Id']
    try:
        return api_u.render(
            clusters=[c.dict for c in api.get_clusters(tenant_id=tenant_id)])
    except Exception, e:
        return api_u.internal_error(500, 'Exception while listing Clusters', e)


@rest.post('/clusters')
@v.validate(v.validate_cluster_create)
def clusters_create():
    data = api_u.request_data()
    headers = request.headers

    return api_u.render(api.create_cluster(data, headers).wrapped_dict)


@rest.get('/clusters/<cluster_id>')
@v.exists_by_id(api.get_cluster, 'cluster_id', tenant_specific=True)
def clusters_get(cluster_id):
    tenant_id = request.headers['X-Tenant-Id']
    c = api.get_cluster(id=cluster_id, tenant_id=tenant_id)
    return api_u.render(c.wrapped_dict)


@rest.put('/clusters/<cluster_id>')
def clusters_update(cluster_id):
    return api_u.internal_error(501, NotImplementedError(
        "Cluster update op isn't implemented (id '%s')"
        % cluster_id))


@rest.delete('/clusters/<cluster_id>')
@v.exists_by_id(api.get_cluster, 'cluster_id', tenant_specific=True)
def clusters_delete(cluster_id):
    headers = request.headers
    tenant_id = headers['X-Tenant-Id']
    api.terminate_cluster(headers, id=cluster_id, tenant_id=tenant_id)

    return api_u.render()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import request
import functools
import jsonschema
from oslo.config import cfg

from savanna import exceptions as ex
import savanna.openstack.common.exception as os_ex
from savanna.openstack.common import log as logging
from savanna.service import api
import savanna.utils.api as api_u
from savanna.utils.openstack import nova


LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_opt('allow_cluster_ops', 'savanna.config')

# Base validation schema of cluster creation operation
CLUSTER_CREATE_SCHEMA = {
    "title": "Cluster creation schema",
    "type": "object",
    "properties": {
        "cluster": {
            "type": "object",
            "properties": {
                "name": {"type": "string",
                         "minLength": 1,
                         "maxLength": 50,
                         "pattern": r"^(([a-zA-Z]|[a-zA-Z][a-zA-Z0-9\-]"
                                    r"*[a-zA-Z0-9])\.)*([A-Za-z]|[A-Za-z]"
                                    r"[A-Za-z0-9\-]*[A-Za-z0-9])$"},
                "base_image_id": {"type": "string",
                                  "minLength": 1,
                                  "maxLength": 240},
                "node_templates": {
                    "type": "object"
                }
            },
            "required": ["name", "base_image_id", "node_templates"]
        }
    },
    "required": ["cluster"]
}

# Base validation schema of node template creation operation
TEMPLATE_CREATE_SCHEMA = {
    "title": "Node Template creation schema",
    "type": "object",
    "properties": {
        "node_template": {
            "type": "object",
            "properties": {
                "name": {"type": "string",
                         "minLength": 1,
                         "maxLength": 240,
                         "pattern": r"^(([a-zA-Z]|[a-zA-Z][a-zA-Z0-9\-_]"
                                    r"*[a-zA-Z0-9])\.)*([A-Za-z]|[A-Za-z]"
                                    r"[A-Za-z0-9\-_]*[A-Za-z0-9])$"},
                "node_type": {"type": "string",
                              "minLength": 1,
                              "maxLength": 240},
                "flavor_id": {"type": "string",
                              "minLength": 1,
                              "maxLength": 240},
                "task_tracker": {
                    "type": "object"
                },
                "job_tracker": {
                    "type": "object"
                },
                "name_node": {
                    "type": "object"
                },
                "data_node": {
                    "type": "object"
                }
            },
            "required": ["name", "node_type", "flavor_id"]
        }
    },
    "required": ["node_template"]
}


def validate(validate_func):
    def decorator(func):
        @functools.wraps(func)
        def handler(*args, **kwargs):
            try:
                validate_func(api_u.request_data(), **kwargs)
            except jsonschema.ValidationError, e:
                e.code = "VALIDATION_ERROR"
                return api_u.bad_request(e)
            except ex.SavannaException, e:
                return api_u.bad_request(e)
            except os_ex.MalformedRequestBody, e:
                e.code = "MALFORMED_REQUEST_BODY"
                return api_u.bad_request(e)
            except Exception, e:
                return api_u.internal_error(
                    500, "Error occurred during validation", e)

            return func(*args, **kwargs)

        return handler

    return decorator


def exists_by_id(service_func, id_prop, tenant_specific=False):
    def decorator(func):
        @functools.wraps(func)
        def handler(*args, **kwargs):
            try:
                if tenant_specific:
                    tenant = request.headers['X-Tenant-Id']
                    service_func(*args, id=kwargs[id_prop], tenant_id=tenant)
                else:
                    service_func(*args, id=kwargs[id_prop])
                return func(*args, **kwargs)
            except ex.NotFoundException, e:
                e.__init__(kwargs[id_prop])
                return api_u.not_found(e)
            except Exception, e:
                return api_u.internal_error(
                    500, "Unexpected error occurred", e)

        return handler

    return decorator


def validate_cluster_create(cluster_values):
    jsonschema.validate(cluster_values, CLUSTER_CREATE_SCHEMA)
    values = cluster_values['cluster']

    # check that requested cluster name is unique
    unique_names = [cluster.name for cluster in api.get_clusters()]
    if values['name'] in unique_names:
        raise ex.ClusterNameExistedException(values['name'])

    # check that requested templates are from already defined values
    node_templates = values['node_templates']
    possible_node_templates = [nt.name for nt in api.get_node_templates()]
    for nt in node_templates:
        if nt not in possible_node_templates:
            raise ex.NodeTemplateNotFoundException(nt)
        # check node count is integer and non-zero value
        jsonschema.validate(node_templates[nt],
                            {"type": "integer", "minimum": 1})

    # check that requested cluster contains only 1 instance of NameNode
    # and 1 instance of JobTracker
    jt_count = 0
    nn_count = 0

    for nt_name in node_templates:
        processes = api.get_node_template(name=nt_name).dict['node_type'][
            'processes']
        if "job_tracker" in processes:
            jt_count += node_templates[nt_name]
        if "name_node" in processes:
            nn_count += node_templates[nt_name]

    if nn_count != 1:
        raise ex.NotSingleNameNodeException(nn_count)

    if jt_count != 1:
        raise ex.NotSingleJobTrackerException(jt_count)

    if CONF.allow_cluster_ops:
        image_id = values['base_image_id']
        nova_images = nova.get_images(request.headers)
        if image_id not in nova_images:
            LOG.debug("Could not find %s image in %s", image_id, nova_images)
            raise ex.ImageNotFoundException(values['base_image_id'])

        # check available Nova absolute limits
        _check_limits(nova.get_limits(request.headers),
                      values['node_templates'])
    else:
        LOG.info("Cluster ops are disabled, use --allow-cluster-ops flag")


def validate_node_template_create(nt_values):
    jsonschema.validate(nt_values, TEMPLATE_CREATE_SCHEMA)
    values = nt_values['node_template']

    # check that requested node_template name is unique
    unique_names = [nt.name for nt in api.get_node_templates()]
    if values['name'] in unique_names:
        raise ex.NodeTemplateExistedException(values['name'])

    node_types = [nt.name for nt in api.get_node_types()]

    if values['node_type'] not in node_types:
        raise ex.NodeTypeNotFoundException(values['node_type'])

    req_procs = []
    if "TT" in values['node_type']:
        req_procs.append("task_tracker")
    if "DN" in values['node_type']:
        req_procs.append("data_node")
    if "NN" in values['node_type']:
        req_procs.append("name_node")
    if "JT" in values['node_type']:
        req_procs.append("job_tracker")

    LOG.debug("Required properties are: %s", req_procs)

    jsonschema.validate(values, {"required": req_procs})

    processes = values.copy()
    del processes['name']
    del processes['node_type']
    del processes['flavor_id']

    LOG.debug("Incoming properties are: %s", processes)

    for proc in processes:
        if proc not in req_procs:
            raise ex.DiscrepancyNodeProcessException(req_procs)

    req_params = api.get_node_type_required_params(name=values['node_type'])
    for process in req_params:
        for param in req_params[process]:
            if param not in values[process] or not values[process][param]:
                raise ex.RequiredParamMissedException(process, param)

    all_params = api.get_node_type_all_params(name=values['node_type'])
    for process in all_params:
        for param in processes[process]:
            if param not in all_params[process]:
                raise ex.ParamNotAllowedException(param, process)

    if api.CONF.allow_cluster_ops:
        flavor = values['flavor_id']
        nova_flavors = nova.get_flavors(request.headers)
        if flavor not in nova_flavors:
            LOG.debug("Could not find %s flavor in %s", flavor, nova_flavors)
            raise ex.FlavorNotFoundException(flavor)
    else:
        LOG.info("Cluster ops are disabled, use --allow-cluster-ops flag")


def _check_limits(limits, node_templates):
    all_vcpus = limits['maxTotalCores'] - limits['totalCoresUsed']
    all_ram = limits['maxTotalRAMSize'] - limits['totalRAMUsed']
    all_inst = limits['maxTotalInstances'] - limits['totalInstancesUsed']
    LOG.info("List of available VCPUs: %d, RAM: %d, Instances: %d",
             all_vcpus, all_ram, all_inst)

    need_vcpus = 0
    need_ram = 0
    need_inst = 0
    for nt_name in node_templates:
        nt_flavor_name = api.get_node_template(name=nt_name).dict['flavor_id']
        nt_flavor_count = node_templates[nt_name]
        LOG.debug("User requested flavor: %s, count: %s",
                  nt_flavor_name, nt_flavor_count)
        nova_flavor = nova.get_flavor(request.headers, name=nt_flavor_name)
        LOG.debug("Nova has flavor %s with VCPUs=%d, RAM=%d",
                  nova_flavor.name, nova_flavor.vcpus, nova_flavor.ram)

        need_vcpus += nova_flavor.vcpus * nt_flavor_count
        need_ram += nova_flavor.ram * nt_flavor_count
        need_inst += nt_flavor_count

    LOG.info("User requested %d instances with total VCPUs=%d and RAM=%d",
             need_inst, need_vcpus, need_ram)

    if need_inst > all_inst or need_vcpus > all_vcpus or need_ram > all_ram:
        raise ex.NotEnoughResourcesException([all_inst, all_vcpus, all_ram,
                                              need_inst, need_vcpus, need_ram])


def validate_node_template_terminate(_, template_id):
    if api.is_node_template_associated(id=template_id):
        name = api.get_node_template(id=template_id).name
        raise ex.AssociatedNodeTemplateTerminationException(name)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna.tests.integration.db import ITestCase
import savanna.tests.integration.parameters as param
from telnetlib import Telnet


class ITestClusterApi(ITestCase):

    def setUp(self):
        super(ITestClusterApi, self).setUp()
        Telnet(self.host, self.port)

    def test_cluster_crud_operations(self):
        nt_body = self.make_nt('master-node', 'JT+NN', 1234, 2345)
        data_nt_master = self._post_object(self.url_nt, nt_body, 202)

        nt_body = self.make_nt('worker-node', 'TT+DN', 1234, 2345)
        data_nt_worker = self._post_object(self.url_nt, nt_body, 202)

        try:
            cluster_body = self.make_cluster_body(
                param.CLUSTER_NAME_CRUD, 'master-node', 'worker-node', 2)
            get_cluster_body = self._get_body_cluster(
                param.CLUSTER_NAME_CRUD, 'master-node', 'worker-node', 2)

            self._crud_object(cluster_body, get_cluster_body, self.url_cluster)

        finally:
            self.delete_node_template(data_nt_master)
            self.delete_node_template(data_nt_worker)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from novaclient import client as nc
from os import getcwd
import paramiko
from re import search
from savanna.service.cluster_ops import _setup_ssh_connection
from savanna.tests.integration.db import ITestCase
import savanna.tests.integration.parameters as param
from telnetlib import Telnet


def _open_transport_chanel(transport):
    transport.connect(
        username=param.NODE_USERNAME, password=param.NODE_PASSWORD)
    return paramiko.SFTPClient.from_transport(transport)


def _execute_transfer_to_node(host, locfile, nodefile):
    try:
        transport = paramiko.Transport(host)
        sftp = _open_transport_chanel(transport)
        sftp.put(locfile, nodefile)

    finally:
        sftp.close()
        transport.close()


def _execute_transfer_from_node(host, nodefile, localfile):
    try:
        transport = paramiko.Transport(host)
        sftp = _open_transport_chanel(transport)
        sftp.get(nodefile, localfile)

    finally:
        sftp.close()
        transport.close()


def _open_channel_and_execute(ssh, cmd, print_output):
    chan = ssh.get_transport().open_session()
    chan.exec_command(cmd)
    stdout = chan.makefile('rb', -1)
    chan.set_combine_stderr(True)
    if print_output:
        return stdout.read()
    return chan.recv_exit_status()


def _execute_command_on_node(host, cmd, print_output=False):
    ssh = paramiko.SSHClient()
    try:
        _setup_ssh_connection(host, ssh)
        return _open_channel_and_execute(ssh, cmd, print_output)
    finally:
        ssh.close()


def _transfer_script_to_node(host, directory):
    _execute_transfer_to_node(
        str(host), '%s/integration/script.sh' % directory, 'script.sh')
    _execute_command_on_node(str(host), 'chmod 777 script.sh')


class TestHadoop(ITestCase):

    def setUp(self):
        super(TestHadoop, self).setUp()
        Telnet(self.host, self.port)

    def _hadoop_testing(self, cluster_name, nt_name_master,
                        nt_name_worker, number_workers):
        object_id = None
        cluster_body = self.make_cluster_body(
            cluster_name, nt_name_master, nt_name_worker, number_workers)
        data = self._post_object(self.url_cluster, cluster_body, 202)

        try:
            data = data['cluster']
            object_id = data.pop(u'id')
            get_body = self._get_body_cluster(
                cluster_name, nt_name_master, nt_name_worker, number_workers)
            get_data = self._get_object(self.url_cl_with_slash, object_id, 200)
            get_data = get_data['cluster']
            del get_data[u'id']
            self._await_cluster_active(
                get_body, get_data, self.url_cl_with_slash, object_id)

            get_data = self._get_object(
                self.url_cl_with_slash, object_id, 200, True)
            get_data = get_data['cluster']
            namenode = get_data[u'service_urls'][u'namenode']
            jobtracker = get_data[u'service_urls'][u'jobtracker']
            nodes = get_data[u'nodes']
            worker_ips = []
            nova = nc.Client(version='2',
                             username=param.OS_USERNAME,
                             api_key=param.OS_PASSWORD,
                             auth_url=param.OS_AUTH_URL,
                             project_id=param.OS_TENANT_NAME)
            for node in nodes:
                if node[u'node_template'][u'name'] == '%s.%s'\
                        % (nt_name_worker, param.FLAVOR_ID):
                    v = nova.servers.get('%s' % node[u'vm_id'])
                    for network, address in v.addresses.items():
                        instance_ips = json.dumps(address)
                        instance_ips = json.loads(instance_ips)
                    for instance_ip in instance_ips:
                        if instance_ip[u'addr'][:len(param.IP_PREFIX)]\
                                == param.IP_PREFIX:
                            worker_ips.append(instance_ip[u'addr'])

            p = '(?:http.*://)?(?P<host>[^:/ ]+).?(?P<port>[0-9]*).*'
            m = search(p, namenode)
            t = search(p, jobtracker)

            namenode_ip = m.group('host')
            namenode_port = m.group('port')
            jobtracker_ip = t.group('host')
            jobtracker_port = t.group('port')

            try:
                Telnet(str(namenode_ip), str(namenode_port))
                Telnet(str(jobtracker_ip), str(jobtracker_port))
            except Exception as e:
                self.fail('telnet nn or jt is failure: ' + e.message)

            this_dir = getcwd()

            try:
                _transfer_script_to_node(namenode_ip, this_dir)
                for worker_ip in worker_ips:
                    _transfer_script_to_node(worker_ip, this_dir)
            except Exception as e:
                self.fail('failure in transfer script: ' + e.message)

            try:
                self.assertEqual(int(_execute_command_on_node(
                    namenode_ip, './script.sh lt -hd %s'
                                 % param.HADOOP_DIRECTORY, True)),
                                 number_workers)
                #TODO(vrovachev) delete sleep from script after fix bug 1183387
            except Exception as e:
                self.fail('compare number active trackers is failure: '
                          + e.message)

            try:
                self.assertEqual(int(_execute_command_on_node(
                    namenode_ip, './script.sh ld -hd %s'
                                 % param.HADOOP_DIRECTORY, True)),
                                 number_workers)
            except Exception as e:
                self.fail('compare number active datanodes is failure: '
                          + e.message)

            try:
                _execute_command_on_node(
                    namenode_ip, './script.sh pi -nc %s -hv %s -hd %s'
                                 % (number_workers, param.HADOOP_VERSION,
                                    param.HADOOP_DIRECTORY))
            except Exception as e:
                _execute_transfer_from_node(
                    namenode_ip,
                    '/outputTestMapReduce/log.txt', '%s/errorLog' % this_dir)
                self.fail(
                    'run pi script or get run in active trackers is failure: '
                    + e.message)

            try:
                job_name = _execute_command_on_node(
                    namenode_ip, './script.sh gn -hd %s'
                                 % param.HADOOP_DIRECTORY, True)
            except Exception as e:
                self.fail('fail in get job name: ' + e.message)

            try:
                for worker_ip in worker_ips:
                    self.assertEquals(
                        _execute_command_on_node(
                            worker_ip,
                            './script.sh ed -jn %s -hld %s'
                            % (job_name[:-1], param.HADOOP_LOG_DIRECTORY)), 0)
            except Exception as e:
                self.fail('fail in check run job in worker nodes: '
                          + e.message)

            try:
                self.assertEquals(
                    _execute_command_on_node(
                        namenode_ip, './script.sh mr -hv %s -hd %s'
                                     % (param.HADOOP_VERSION,
                                        param.HADOOP_DIRECTORY)), 0)
            except Exception as e:
                _execute_transfer_from_node(
                    namenode_ip,
                    '/outputTestMapReduce/log.txt', '%s/errorLog' % this_dir)
                self.fail('run hdfs script is failure: ' + e.message)
        except Exception as e:
            self.fail(e.message)

        finally:
            self._del_object(self.url_cl_with_slash, object_id, 204)

    def test_hadoop_single_master(self):
        data_nt_master = self._post_object(
            self.url_nt, self.make_nt('master_node', 'JT+NN',
                                      1234, 1234), 202)
        data_nt_worker = self._post_object(
            self.url_nt, self.make_nt('worker_node', 'TT+DN',
                                      1234, 1234), 202)

        try:
            self._hadoop_testing(
                param.CLUSTER_NAME_HADOOP, 'master_node', 'worker_node', 2)
        except Exception as e:
            self.fail(e.message)

        finally:
            self.delete_node_template(data_nt_master)
            self.delete_node_template(data_nt_worker)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna.service.cluster_ops import _create_xml
import unittest


class ConfigGeneratorTest(unittest.TestCase):
    def test_xml_generator(self):
        config = {
            'key-1': 'value-1',
            'key-2': 'value-2',
            'key-3': 'value-3',
            'key-4': 'value-4',
            'key-5': 'value-5',
        }
        xml = _create_xml(config, config.keys())
        self.assertEqual(xml, """<?xml version="1.0" ?>
<?xml-stylesheet type="text/xsl" href="configuration.xsl"?>
<configuration>
  <property>
    <name>key-3</name>
    <value>value-3</value>
  </property>
  <property>
    <name>key-2</name>
    <value>value-2</value>
  </property>
  <property>
    <name>key-1</name>
    <value>value-1</value>
  </property>
  <property>
    <name>key-5</name>
    <value>value-5</value>
  </property>
  <property>
    <name>key-4</name>
    <value>value-4</value>
  </property>
</configuration>
""")
